###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

import os
import re
import types
import fnmatch

import cherrypy
import simplejson

from mako.template import Template
from mako.lookup import TemplateLookup

from openerp import rpc
from openerp.tools import utils


__all__ = ['find_resource', 'load_template', 'render_template', 'expose', 'register_template_vars']


def find_resource(package_or_module, *names):

    ref = package_or_module
    if isinstance(package_or_module, basestring):
        ref = __import__(package_or_module, globals(), \
                fromlist=package_or_module.split('.'))

    return os.path.abspath(os.path.join(os.path.dirname(ref.__file__), *names))


def find_resources(package_or_module, *patterns):
    root = find_resource("openerp")
    for path, dirs, files in os.walk(os.path.abspath(root)):
        for pattern in patterns:
            for filename in fnmatch.filter(files, pattern):
                yield os.path.join(path, filename)
                
                
# ask autoreloader to check mako templates and cfg files
for res in find_resources("openerp", "*.mako", "*.cfg"):
    cherrypy.engine.autoreload.files.add(res)


filters = ["__content"]
imports = ["from openerp.tools.utils import content as __content"]

class TL(TemplateLookup):
    
    cache = {}
    
    def get_template(self, uri):
        try:
            return self.cache[str(uri)]
        except Exception, e:
            pass
        self.cache[str(uri)] = res = super(TL, self).get_template(uri)
        return res
    
template_lookup = TL(directories=[find_resource("openerp")], 
                        default_filters=filters,
                        imports=imports)#, module_directory="mako_modules")

def load_template(template, module=None):

    if not isinstance(template, basestring):
        return template
    
    if re.match('(.+)\.(html|mako)\s*$', template):

        if module:
            template = find_resource(module, template)
        else:
            template = os.path.abspath(template)
            
        base = find_resource("openerp")
        template = template.replace(base, '').replace('\\', '/')
        
        return template_lookup.get_template(template)

    else:
        return Template(template, default_filters=filters, imports=imports)


class _Provider(dict):

    def __getattr__(self, name):
        if name in self:
            return self[name]
        return super(_Provider, self).__getattribute__(name)

    def __setattr__(self, name, value):
        raise AttributeError


_var_providers = {}


def register_template_vars(callback, prefix='oo'):
    providers = _var_providers.setdefault(prefix, [])
    providers.append(callback)


def _cp_vars():

    return {
        'session': cherrypy.session,
        'request': cherrypy.request,
        'config': utils.config,
        'root': cherrypy.request.app.root,
    }


def _py_vars():

    return {
        'url': utils.url,
        'attrs': utils.attrs,
        'attr_if': utils.attr_if,
        'checker': lambda e: utils.attr_if('checked', e),
        'selector': lambda e: utils.attr_if('selected', e),
        'readonly': lambda e: utils.attr_if('readonly', e),
        'disabled': lambda e: utils.attr_if('disabled', e),
    }


def _root_vars():
    return {
        'rpc': rpc,
    }


register_template_vars(_cp_vars, 'cp')
register_template_vars(_py_vars, 'py')
register_template_vars(_root_vars, None)

def _get_vars():
    
    try:
        return cherrypy.request._terp_template_vars
    except:
        pass
    
    cherrypy.request._terp_template_vars = _vars = {}
    
    for prefix, cbs in _var_providers.iteritems():
        if prefix:
            provider = _Provider()
            for cb in cbs:
                provider.update(cb())
            _vars[prefix] = provider
        else:
            for cb in cbs:
                _vars.update(cb())
    return _vars

def render_template(template, kw):
    
    assert isinstance(template, Template), "Invalid template..."
    
    kw.update(_get_vars())
    
    # XXX mako overrides 'context' template variable...
    if 'context' in kw:
        kw['ctx'] = kw.pop('context')
        
    return utils.NoEscape(template.render_unicode(**kw))


def expose(format='html', template=None, content_type=None, allow_json=False):

    def expose_wrapper(func):
        
        template_c = load_template(template, func.__module__)

        def func_wrapper(*args, **kw):

            res = func(*args, **kw)
            
            if format == 'json' or (allow_json and 'allow_json' in cherrypy.request.params):
                cherrypy.response.headers['Content-Type'] = 'text/javascript'
                return simplejson.dumps(res)
            
            cherrypy.response.headers['Content-Type'] = content_type or \
            cherrypy.response.headers.get('Content-Type', 'text/html')
            
            if isinstance(res, dict):
                
                try:
                    _template = load_template(res['cp_template'], func.__module__)
                except:
                    _template = template_c
                    
                if _template:
                    
                    from openerp.widgets import Widget, OrderedSet
                    from openerp.widgets import js_i18n
                    
                    res['widget_css'] = css = OrderedSet()
                    res['widget_javascript'] = js = {}
                    
                    jset = js.setdefault('head', OrderedSet())
                    jset.add_all(js_i18n.retrieve_javascript())
                                        
                    for value in res.itervalues():
                        
                        if isinstance(value, Widget):
                            css.add_all(value.retrieve_css())
                            for script in value.retrieve_javascript():
                                jset = js.setdefault(script.location or 'head', OrderedSet())
                                jset.add(script)

                    return render_template(_template, res).encode("utf-8")
                
            if not isinstance(res, basestring):
                return unicode(res).encode("utf-8")

            return res

        return utils.decorated(func_wrapper, func, exposed=True)

    return expose_wrapper


# vim: ts=4 sts=4 sw=4 si et




from openerp.widgets.base import *
from openerp.widgets.interface import *

import form
import graph
import listgrid
import treegrid

import screen
import form_view

from js_i18n import js_i18n

# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

import math
import time
import datetime
import xml.dom.minidom

import cherrypy

from openerp import rpc
from openerp import tools

from openerp.i18n import format
from openerp.utils import TinyDict

from openerp.widgets.interface import TinyWidget
from openerp.widgets.interface import ConcurrencyInfo

from openerp.widgets import JSLink, CSSLink

from utils import Day
from utils import parse_datetime

COLOR_PALETTE = ['#f57900', '#cc0000', '#d400a8', '#75507b', '#3465a4', '#73d216', '#c17d11', '#edd400',
                 '#fcaf3e', '#ef2929', '#ff00c9', '#ad7fa8', '#729fcf', '#8ae234', '#e9b96e', '#fce94f',
                 '#ff8e00', '#ff0000', '#b0008c', '#9000ff', '#0078ff', '#00ff00', '#e6ff00', '#ffff00',
                 '#905000', '#9b0000', '#840067', '#510090', '#0000c9', '#009b00', '#9abe00', '#ffc900',]

_colorline = ['#%02x%02x%02x' % (25+((r+10)%11)*23,5+((g+1)%11)*20,25+((b+4)%11)*23) for r in range(11) for g in range(11) for b in range(11) ]
def choice_colors(n):
    if n > len(COLOR_PALETTE):
        return _colorline[0:-1:len(_colorline)/(n+1)]
    elif n:
        return COLOR_PALETTE[:n]
    return []

class TinyEvent(TinyWidget):

    template = """<div class="calEvent" style="background-color: ${color}"
    starts="${str(starts)}" ends="${str(ends)}" record_id="${record_id}" title="${description}">${title}</div>
    """

    params = ['starts', 'ends', 'title', 'description', 'color', 'record_id']

    starts = None
    ends = None
    dayspan = 0
    color = None

    title = ''
    description = ''

    record = {}
    record_id = False

    def __init__(self, record, starts, ends, title='', description='', dayspan=0, color=None):

        super(TinyEvent, self).__init__()

        self.record = record
        self.record_id = record['id']

        if starts and ends:

            self.starts = (starts or None) and datetime.datetime(*starts[:6])
            self.ends = (ends or None) and datetime.datetime(*ends[:6])

        self.dayspan = dayspan

        self.title = title
        self.description = description
        self.color = color

class ICalendar(TinyWidget):
    """ Base Calendar calss
    """

    mode = 'month'
    date_start = None
    date_delay = None
    date_stop = None
    color_field = None
    day_length = 8
    use_search = False
    selected_day = None
    date_format = '%Y-%m-%d'

    params = ['use_search']
    member_widgets = ['concurrency_info']

    css = [CSSLink('openerp', 'tinycalendar/css/calendar.css')]
    javascript = [JSLink('openerp', 'tinycalendar/javascript/calendar_date.js'),
                  JSLink('openerp', 'tinycalendar/javascript/calendar_utils.js'),
                  JSLink('openerp', 'tinycalendar/javascript/calendar_box.js'),
                  JSLink('openerp', 'tinycalendar/javascript/calendar_month.js'),
                  JSLink('openerp', 'tinycalendar/javascript/calendar_week.js')]

    def __init__(self, model, ids, view, domain=[], context={}, options=None):

        super(ICalendar, self).__init__()

        self.info_fields = []
        self.fields = []

        self.events = {}

        self.colors = {}
        self.color_values = []

        self.calendar_fields = {}
        self.concurrency_info = None

        self.ids = ids
        self.model = model
        self.domain = domain or []
        self.context = context or {}
        self.options = options

        self.date_format = format.get_datetime_format('date')
        self.use_search = (options or None) and options.use_search

        try:
            dt = parse_datetime(options.selected_day)
            self.selected_day = Day(dt.year, dt.month, dt.day)
        except:
            pass

        proxy = rpc.RPCProxy(model)

        view_id = view.get('view_id', False)

        dom = xml.dom.minidom.parseString(view['arch'].encode('utf-8'))
        root = dom.childNodes[0]
        attrs = tools.node_attributes(root)

        self.string = attrs.get('string', '')
        self.date_start = attrs.get('date_start')
        self.date_delay = attrs.get('date_delay')
        self.date_stop = attrs.get('date_stop')
        self.color_field = attrs.get('color')
        self.day_length = int(attrs.get('day_length', 8))

        if options and options.mode:
            self.mode = options.mode
        else:
            self.mode = attrs.get('mode') or self.mode or 'month'

        self.info_fields = self.parse(root, view['fields'])

        fields = view['fields']
        fields = fields.keys() + [self.date_start, self.date_stop, self.date_delay, self.color_field]

        fields = list(set([x for x in fields if x]))

        self.fields = proxy.fields_get(fields)

        if self.color_field and options and options.colors:
            self.colors = options.colors

        if self.color_field and options and options.color_values:
            self.color_values = options.color_values

        self.calendar_fields['date_start'] = dict(name=self.date_start,
                                                  kind=self.fields[self.date_start]['type'])

        if self.date_delay:
            self.calendar_fields['date_delay'] = dict(name=self.date_delay,
                                                      kind=self.fields[self.date_delay]['type'])

        if self.date_stop:
            self.calendar_fields['date_stop'] = dict(name=self.date_stop,
                                                         kind=self.fields[self.date_stop]['type'])

        self.calendar_fields['day_length'] = self.day_length

    def parse(self, root, fields):
        """ Deraived class must override parse method
        """
        pass

    def convert(self, event):

        fields = [x for x in [self.date_start, self.date_stop] if x]

        for fld in fields:
            typ = self.fields[fld]['type']
            assert typ in ('date', 'datetime'), "Invalid field type (%s), should be either `date` or `datetime`: %s" % (typ, fld)
            fmt = format.DT_SERVER_FORMATS[typ]

            if event[fld] and fmt:
                event[fld] = time.strptime(event[fld], fmt)

            # default start time is 9:00 AM
            if typ == 'date' and fld == self.date_start:
                ds = list(event[fld])
                ds[3] = 9
                event[fld] = tuple(ds)

    def get_events(self, days):

        proxy = rpc.RPCProxy(self.model)

        #XXX: how to get events not falling between the given day range but spans the range?
        #domain = self.domain + [(self.date_start, '>', days[0].prev().isoformat()),
        #                        (self.date_start, '<', days[-1].next().isoformat())]

        first = days[0].month2.prev()[0] #HACK: add prev month
        domain = self.domain + [(self.date_start, '>', first.isoformat()),
                                (self.date_start, '<', days[-1].next().isoformat())]

        # convert color values from string to python values
        if self.color_values and self.color_field in self.fields:
            try:
                import openerp.widgets as tw
                atr = self.fields[self.color_field]
                atr['required'] = False
                wid = tw.form.WIDGETS[atr['type']](**atr)
                vals = self.color_values[:]
                for i, v in enumerate(vals):
                    try:
                        vals[i] = wid.validator.to_python(v)
                    except:
                        pass
                domain += [(self.color_field, "in", vals)]
            except Exception, e:
                pass

        if self.options and self.options.use_search:
            domain += self.options.search_domain

        ctx = rpc.session.context.copy()
        ctx.update(self.context)

        order_by = ('sequence' in self.fields or 0) and 'sequence'

        ids = proxy.search(domain, 0, 0, order_by, ctx)
        result = proxy.read(ids, self.fields.keys()+['__last_update'], ctx)
        self._update_concurrency_info(self.model, result)
        self.concurrency_info = ConcurrencyInfo(self.model, ids)

        if self.color_field:

            for evt in result:
                key = evt[self.color_field]
                name = key
                value = key

                if isinstance(key, list): # M2O, XMLRPC returns List instead of Tuple
                    evt[self.color_field] = key = tuple(key)

                if isinstance(key, tuple): # M2O
                    value, name = key

                self.colors[key] = (name, value, None)

            colors = choice_colors(len(self.colors))
            for i, (key, value) in enumerate(self.colors.items()):
                self.colors[key] = (value[0], value[1], colors[i])

        events = []

        for evt in result:
            self.convert(evt)
            events += [self.get_event_widget(evt)]

        # filter out the events which are not in the range
        result = []
        for e in events:
            if e.dayspan > 0 and days[0] - e.dayspan < e.starts:
                result += [e]
            if e.dayspan == 0 and days[0] <= e.starts:
                result += [e]

        return result

    def get_event_widget(self, event):

        title = ''       # the title
        description = [] # the description
        starts = None    # the starting time (datetime)
        ends = None      # the end time (datetime)

        if self.info_fields:

            f = self.info_fields[0]
            s = event[f]

            if isinstance(s, (tuple, list)): s = s[-1]

            title = ustr(s)

            for f in self.info_fields[1:]:
                s = event[f]
                if isinstance(s, (tuple, list)): s = s[-1]

                description += [ustr(s)]

        starts = event.get(self.date_start)
        ends = event.get(self.date_delay) or 1.0
        span = 0

        if starts and ends:

            n = 0
            h = ends

            if ends == self.day_length:
                n += 1

            elif ends > self.day_length:
                n = ends / self.day_length
                h = ends % self.day_length

                n = int(math.floor(n))

                if h > 0: n += 1

            span = n
            ends = time.localtime(time.mktime(starts) + (h * 60 * 60) + (n * 24 * 60 * 60))

        if starts and self.date_stop:

            ends = event.get(self.date_stop)
            if not ends:
                ends = time.localtime(time.mktime(starts) + 60 * 60)

            tds = time.mktime(starts)
            tde = time.mktime(ends)

            if tds >= tde:
                tde = tds + 60 * 60
                ends = time.localtime(tde)

            n = (tde - tds) / (60 * 60)

            if n >= self.day_length:
                span = math.ceil(n / 24)

        starts = format.format_datetime(starts, "datetime", True)
        ends = format.format_datetime(ends, "datetime", True)

        color_key = event.get(self.color_field)
        color = self.colors.get(color_key)

        title = title.strip()
        description = ', '.join(description).strip()

        return TinyEvent(event, starts, ends, title, description, dayspan=span, color=(color or None) and color[-1])


class TinyCalendar(ICalendar):

    def parse(self, root, fields):

        info_fields = []
        attrs = tools.node_attributes(root)

        for node in root.childNodes:
            attrs = tools.node_attributes(node)

            if node.localName == 'field':
                info_fields += [attrs['name']]

        return info_fields

# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

import cherrypy

from openerp.tools import expose

from openerp import rpc
from openerp import tools
from openerp.controllers.base import SecuredController

from openerp.utils import TinyDict
from openerp.utils import TinyForm
from openerp.utils import TinyFormError

import openerp.widgets as tw

import form
import search
import wizard

class List(SecuredController):

    @expose('json')
    def save(self, **kw):
        params, data = TinyDict.split(kw)

        error = None
        error_field = None

        id = params.id or 0
        id = (id > 0) and id or 0

        ids = params.ids or []

        model = params.parent.model
        if model != params.model and not params.parent.id:
            error = _("Parent record doesn't exists...")

        if error:
            return dict(error=error)

        try:
            proxy = rpc.RPCProxy(model)
            frm = TinyForm(**kw).to_python()
            data = {}
            ctx = tools.context_with_concurrency_info(params.parent.context, params.concurrency_info)

            if model != params.model:
                source = params.source
                data = frm.chain_get(source)

                if '__id' in data: data.pop('__id')
                if 'id' in data: data.pop('id')

                fld = source.split('/')[-1]
                data = {fld : [(id and 1, id, data.copy())]}
                proxy.write([params.parent.id], data, ctx)

                all_ids = proxy.read([params.parent.id], [fld])[0][fld]
                new_ids = [i for i in all_ids if i not in ids]

                ids = all_ids
                if new_ids:
                    id = new_ids[0]

            else:
                data = frm.copy()
                if 'id' in data: data.pop('id')

                if id > 0:
                    proxy.write([id], data, ctx)
                else:
                    id = proxy.create(data, params.parent.context or {})
                    ids = [id] + ids

        except TinyFormError, e:
            error_field = e.field
            error = ustr(e)
        except Exception, e:
            error = ustr(e)

        return dict(error_field=error_field, error=error, id=id, ids=str([int(i) for i in ids]))

    @expose('json')
    def remove(self, **kw):
        params, data = TinyDict.split(kw)
        error = None
        proxy = rpc.RPCProxy(params.model)
        if params.ids:
            try:
                ctx = tools.context_with_concurrency_info(params.context, params.concurrency_info)
                if isinstance(params.ids, list):
                    res = proxy.unlink(params.ids, ctx)
                else:
                    res = proxy.unlink([params.ids], ctx)
            except Exception, e:
                error = ustr(e)

        return dict(error=error)

    @expose('json')
    def get(self, **kw):
        params, data = TinyDict.split(kw)

        params.ids = None
        source = (params.source or '') and str(params.source)

        params.view_type = 'form'

        if source == '_terp_list':
            params.view_type = 'tree'
            if params.search_domain:
                params.domain += params.search_domain

        # default_get context
        current = params.chain_get(source)
        if current and params.source_default_get:
            current.context = current.context or {}
            current.context.update(params.source_default_get)

        if params.wiz_id:
            res = wizard.Wizard().execute(params)
            frm = res['form']
        else:
            frm = form.Form().create_form(params)

        wid = frm.screen.get_widgets_by_name(source, kind=tw.listgrid.List)[0]
        ids = wid.ids
        count = wid.count

        if params.edit_inline:
            wid.edit_inline = params.edit_inline

        info = {}
        if params.concurrency_info:
            for m, v in getattr(cherrypy.request, 'terp_concurrency_info', {}).items():
                for i, d in v.items():
                    info['%s,%s' % (m, i)] = d

        return dict(ids=ids, count=count, view=ustr(wid.render()), info=info)

    @expose('json')
    def button_action(self, **kw):
        params, data = TinyDict.split(kw)

        error = None
        reload = (params.context or {}).get('reload', False)
        result = {}

        name = params.button_name
        btype = params.button_type

        id = params.id
        model = params.model

        id = (id or False) and int(id)
        ids = (id or []) and [id]

        try:

            if btype == 'workflow':
                rpc.session.execute('object', 'exec_workflow', model, name, id)

            elif btype == 'object':
                ctx = params.context or {}
                ctx.update(rpc.session.context.copy())
                res = rpc.session.execute('object', 'execute', model, name, ids, ctx)
                if isinstance(res, dict) and res.get('type') == 'ir.actions.act_url':
                    result = res

            elif btype == 'action':
                from openerp.controllers import actions

                action_id = int(name)
                action_type = actions.get_action_type(action_id)

                if action_type == 'ir.actions.wizard':
                    cherrypy.session['wizard_parent_form'] = '/form'
                    cherrypy.session['wizard_parent_params'] = params

                res = actions.execute_by_id(action_id, type=action_type, model=model, id=id, ids=ids)
                if isinstance(res, dict) and res.get('type') == 'ir.actions.act_url':
                    result = res
                elif res:
                    error = "Button action has returned another view..."

            else:
                error = "Unallowed button type"
        except Exception, e:
            error = ustr(e)

        return dict(error=error, result=result, reload=reload)

    @expose('json')
    def moveUp(self, **kw):

        params, data = TinyDict.split(kw)

        id = params.id
        ids = params.ids or []

        if id not in ids or ids.index(id) == 0:
            return dict()

        proxy = rpc.RPCProxy(params.model)
        ctx = rpc.session.context.copy()

        prev_id = ids[ids.index(id)-1]

        try:
            res = proxy.read([id, prev_id], ['sequence'], ctx)
            records = dict([(r['id'], r['sequence']) for r in res])
            cur_seq = records[id]
            prev_seq = records[prev_id]

            if cur_seq == prev_seq:
                proxy.write([prev_id], {'sequence': cur_seq + 1}, ctx)
                proxy.write([id], {'sequence': prev_seq}, ctx)
            else:
                proxy.write([id], {'sequence': prev_seq}, ctx)
                proxy.write([prev_id], {'sequence': cur_seq}, ctx)

            return dict()
        except Exception, e:
            return dict(error=str(e))

    @expose('json')
    def moveDown(self, **kw):
        params, data = TinyDict.split(kw)

        id = params.id
        ids = params.ids or []

        if id not in ids or ids.index(id) == len(ids) - 1:
            return dict()

        proxy = rpc.RPCProxy(params.model)
        ctx = rpc.session.context.copy()

        next_id = ids[ids.index(id)+1]

        try:
            res = proxy.read([id, next_id], ['sequence'], ctx)
            records = dict([(r['id'], r['sequence']) for r in res])

            cur_seq = records[id]
            next_seq = records[next_id]

            if cur_seq == next_seq:
                proxy.write([next_id], {'sequence': cur_seq + 1}, ctx)
                proxy.write([id], {'sequence': next_seq}, ctx)
            else:
                proxy.write([id], {'sequence': next_seq}, ctx)
                proxy.write([next_id], {'sequence': cur_seq}, ctx)

            return dict()
        except Exception, e:
            return dict(error=str(e))

# vim: ts=4 sts=4 sw=4 si et




import os
import sys

from openerp import commands

from cherrypy.test import test, helper
test.prefer_parent_path()

import cherrypy


class TestCase(helper.CPWebCase):
    pass

commands.CPSessionWrapper = dict

class CPSessionWrapper(dict):
    
    def __setattr__(self, name, value):
        self[name] = value

    def __getattr__(self, name):
        return self.get(name)

    def __delattr__(self, name):
        if name in self:
            del self[name]

commands.CPSessionWrapper = CPSessionWrapper

def setup_server():
    configfile = commands.get_config_file()
    commands.setup_server(configfile)
    cherrypy.config.update({'environment': 'test_suite'})

def run():
    
    testList = [
        'test_root_controller',
    ]
    
    clp = test.CommandLineParser(testList)
    success = clp.run()
    if clp.interactive:
        print
        raw_input('hit enter')
    sys.exit(success)


if __name__ == '__main__':
    run()



"""
General i18n utility functions.
"""

import cherrypy


__all__ = ['get_locale']


def lang_in_gettext_format(lang):
    if len(lang) > 2:
        country = lang[3:].upper()
        lang = lang[:2] + "_" + country
    return lang


def parse_http_accept_header(accept):
    """Parse an HTTP Accept header (RFC 2616) into a sorted list.

    The quality factors in the header determine the sort order.
    The values can include possible media-range parameters.
    This function can also be used for the Accept-Charset,
    Accept-Encoding and Accept-Language headers.

    """
    if accept is None:
        return []
    items = []
    for item in accept.split(','):
        params = item.split(';')
        for i, param in enumerate(params[1:]):
            param = param.split('=', 1)
            if param[0].strip() == 'q':
                try:
                    q = float(param[1])
                    if not 0 < q <= 1:
                        raise ValueError
                except (IndexError, ValueError):
                    q = 0
                else:
                    item = ';'.join(params[:i+1])
                break
        else:
            q = 1
        if q:
            item = item.strip()
            if item:
                items.append((item, q))
    items.sort(lambda i1, i2: cmp(i2[1], i1[1]))
    return [item[0] for item in items]

def get_accept_languages(accept):
    """Returns a list of languages, by order of preference, based on an
    HTTP Accept-Language string.See W3C RFC 2616
    (http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html) for specification.
    """
    langs = parse_http_accept_header(accept)
    for index, lang in enumerate(langs):
        langs[index] = lang_in_gettext_format(lang)
    return langs

def get_locale(locale=None):
    
    if locale:
        return locale
    
    try:
        from openerp import rpc
        locale = rpc.session.locale
        if locale: return locale
    except:
        pass
    
    try:
        header = cherrypy.request.headers.get("Accept-Language")
        if header:
            accept_languages = get_accept_languages(header)
            if accept_languages:
                return accept_languages[0]
    except AttributeError:
        pass
    
    return "en"





###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

from openerp.tools import expose
from openerp.tools import redirect

import cherrypy

from openerp import rpc
from openerp import tools
from openerp import common

from openerp.controllers.base import SecuredController

class Shortcuts(SecuredController):

    def my(self):

        if not rpc.session.is_logged():
            return []

        sc = cherrypy.session.get('terp_shortcuts', False)
        if not sc:
            proxy = rpc.RPCProxy('ir.ui.view_sc')
            sc = proxy.get_sc(rpc.session.uid, 'ir.ui.menu', rpc.session.context) or []

            cherrypy.session['terp_shortcuts'] = sc

        for x in sc:
            if isinstance(x['res_id'], (list, tuple)):
                x['res_id'] = x['res_id'][0]

        return sc

    def can_create(self):
        return (rpc.session.is_logged() and
                rpc.session.active_id and
                cherrypy.request.path_info == '/tree/open' and
                cherrypy.request.params.get('model') == 'ir.ui.menu')

    @expose()
    def default(self):
        from openerp.controllers import actions

        domain = [('user_id', '=', rpc.session.uid), ('resource', '=', 'ir.ui.menu')]
        return actions.execute_window(False, 'ir.ui.view_sc', res_id=None, domain=domain, view_type='form', mode='tree,form')

    @expose()
    def add(self, id):
        id = int(id)
        proxy = rpc.RPCProxy('ir.ui.view_sc')

        sc = cherrypy.session.get('terp_shortcuts', False)

        if sc:
            for s in sc:
                if s['res_id'] == id:
                    raise redirect('/tree/open', id=id, model='ir.ui.menu')

        name = rpc.RPCProxy('ir.ui.menu').name_get([id], rpc.session.context)[0][1]
        proxy.create({'user_id': rpc.session.uid, 'res_id': id, 'resource': 'ir.ui.menu', 'name': name})

        sc = proxy.get_sc(rpc.session.uid, 'ir.ui.menu', rpc.session.context)
        cherrypy.session['terp_shortcuts'] = sc

        raise redirect('/tree/open', id=id, model='ir.ui.menu')

# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

from base import CSSLink
from interface import TinyWidget


class Pager(TinyWidget):

    template = "templates/pager.mako"
    params = ['offset', 'limit', 'count', 'prev', 'next', 'page_info', 'pager_id']

    css = [CSSLink('openerp', 'css/pager.css')]

    page_info = None
    pager_id = 1

    def __init__(self, id=False, ids=[], offset=0, limit=20, count=0, view_type='tree'):

        super(Pager, self).__init__()

        self.id = id
        self.ids = ids or []
        self.view_type = view_type

        self.offset = offset or 0
        self.limit = limit or 20
        self.count = count or 0

        if len(self.ids) > self.limit:
            self.ids = self.ids[self.offset:]
            self.ids = self.ids[:min(self.limit, len(self.ids))]

        if self.view_type == 'form':

            index = 0
            if self.id in self.ids:
                index = self.offset + self.ids.index(self.id) + 1

            self.page_info = _("[%s/%s]") % (index or '-', self.count)

            self.prev = index > 1
            self.next = index < self.count

        else:
            index = (self.count or 0) and self.offset + 1

            o = self.offset + len(self.ids)
            o = min(self.count, o)

            self.page_info = _("[%s - %s of %s]") % (index, o, self.count)
            self.prev = self.offset > 0
            self.next = self.offset+len(self.ids) < self.count

# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

"""
This modules implements custom authorization logic for the OpenERP Web Client.
"""

import time
import types
import re

import cherrypy
import pkg_resources

from openerp import rpc
from openerp import tools

from openerp.tools import expose
from openerp.tools import redirect


@expose(template="templates/login.mako")
def login(target, db=None, user=None, password=None, action=None, message=None, origArgs={}):

    url = rpc.session.connection_string
    url = str(url[:-1])

    dblist = rpc.session.listdb()
    if dblist == -1:
        dblist = []
        message = _("Could not connect to server!")
        
    dbfilter = cherrypy.request.app.config['openerp-web'].get('dblist.filter')    
    if dbfilter:

        headers = cherrypy.request.headers
        host = headers.get('X-Forwarded-Host', headers.get('Host'))

        base = re.split('\.|:|/', host)[0]

        if dbfilter == 'NONE':
            dblist = dblist

        if dbfilter == 'EXACT':
            base = base
            dblist = [d for d in dblist if d == base]

        if dbfilter == 'UNDERSCORE':
            base = base + '_'
            dblist = [d for d in dblist if d.startswith(base)]

        if dbfilter == 'BOTH':
            dblist = [d for d in dblist if d.startswith(base + '_') or d == base]
            
    info = None
    try:
        info = rpc.session.execute_noauth('common', 'login_message') or ''
    except:
        pass
    return dict(target=target, url=url, dblist=dblist, db=db, user=user, password=password,
            action=action, message=message, origArgs=origArgs, info=info)

def secured(fn):
    """A Decorator to make a SecuredController controller method secured.
    """
    def clear_login_fields(kw={}):

        if not kw.get('login_action'):
            return

        if kw.has_key('db'): del kw['db']
        if kw.has_key('user'): del kw['user']
        if kw.has_key('password'): del kw['password']
        if kw.has_key('login_action'): del kw['login_action']

    def get_orig_args(kw={}):
        if not kw.get('login_action'):
            return kw

        new_kw = kw.copy()

        if new_kw.has_key('db'): del new_kw['db']
        if new_kw.has_key('user'): del new_kw['user']
        if new_kw.has_key('password'): del new_kw['password']
        if new_kw.has_key('login_action'): del new_kw['login_action']

        return new_kw

    def wrapper(*args, **kw):
        """The wrapper function to secure exposed methods
        """

        if rpc.session.is_logged():
            # User is logged in; allow access
            clear_login_fields(kw)
            return fn(*args, **kw)
        else:
            # User isn't logged in yet.

            db = None
            user = None
            password = None
            message = None

            action = kw.get('login_action')

            # get some settings from cookies
            try:
                db = cherrypy.request.cookie['terp_db'].value
                user = cherrypy.request.cookie['terp_user'].value
            except:
                pass

            db = kw.get('db', db)
            user = kw.get('user', user)
            password = kw.get('password', password)

            # See if the user just tried to log in
            if rpc.session.login(db, user, password) <= 0:
                # Bad login attempt
                if action == 'login':
                    message = _("Bad username or password!")

                return login(cherrypy.request.path_info, message=message,
                        db=db, user=user, action=action, origArgs=get_orig_args(kw))

            # Authorized. Set db, user name in cookies
            expiration_time = time.strftime("%a, %d-%b-%Y %H:%M:%S GMT", time.gmtime(time.time() + ( 60 * 60 * 24 * 365 )))
            cherrypy.response.cookie['terp_db'] = db
            cherrypy.response.cookie['terp_user'] = user.encode('utf-8')
            cherrypy.response.cookie['terp_db']['expires'] = expiration_time;
            cherrypy.response.cookie['terp_user']['expires'] = expiration_time;

            # User is now logged in, so show the content
            clear_login_fields(kw)
            return fn(*args, **kw)

    return tools.decorated(wrapper, fn, secured=True)


def unsecured(fn):
    """A Decorator to make a SecuredController controller method unsecured.
    """

    def wrapper(*args, **kw):
        return fn(*args, **kw)

    return tools.decorated(wrapper, fn, secured=False)


class BaseController(object):
    pass


class SecuredController(BaseController):

    def __getattribute__( self, name ):
        value= object.__getattribute__(self, name)

        if isinstance(value, types.MethodType ) and hasattr(value, "exposed") and not (hasattr(value, "secured") and not value.secured):
            return secured(value)

        # Some other property
        return value


# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

from openerp import rpc
from openerp import common

from interface import TinyInputWidget
from form import Form

from openerp import validators

def get_name(model, id):
    id = (id or False) and int(id)
    name = (id or str('')) and str(id)

    if model and id:
        proxy = rpc.RPCProxy(model)

        try:
            name = proxy.name_get([id], rpc.session.context.copy())
            name = name[0][1]
        except common.TinyWarning, e:
            name = _("== Access Denied ==")
        except Exception, e:
            raise e

    return name

class M2O(TinyInputWidget):
    template = "templates/many2one.mako"
    params=['relation', 'text', 'domain', 'context', 'link', 'readonly']

    domain = []
    context = {}
    link = 1

    def __init__(self, **attrs):

        super(M2O, self).__init__(**attrs)
        self.relation = attrs.get('relation', '')

        self.domain = attrs.get('domain', [])
        self.context = attrs.get('context', {})
        self.link = attrs.get('link')
        self.onchange = None # override onchange in js code

        self.validator = validators.many2one()

    def set_value(self, value):

        if value and isinstance(value, (tuple, list)):
            self.default, self.text = value
        else:
            self.default = value
            self.text = get_name(self.relation, self.default)

    def update_params(self, d):
        super(M2O, self).update_params(d)

        if d['value'] and not d['text']:
            d['text'] = get_name(self.relation, d['value'])

# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

"""
This module implementes heirarchical tree view for a tiny model having
    view_type = 'tree'
"""

import time
import xml.dom.minidom

from openerp.tools import expose
from openerp.tools import url

from openerp import rpc
from openerp import icons
from openerp import tools
from openerp import common
from openerp import cache

from openerp.controllers.base import SecuredController
from openerp.widgets import tree_view

from openerp.utils import TinyDict

DT_FORMAT = '%Y-%m-%d'
DHM_FORMAT = '%Y-%m-%d %H:%M:%S'

class Tree(SecuredController):

    @expose(template="templates/tree.mako")
    def create(self, params):

        view_id = (params.view_ids or False) and params.view_ids[0]
        domain = params.domain
        context = params.context

        res_id = params.ids
        model = params.model

        if view_id:
            view_base =  rpc.session.execute('object', 'execute', 'ir.ui.view', 'read', [view_id], ['model', 'type'], context)[0]
            model = view_base['model']
            view = cache.fields_view_get(model, view_id, view_base['type'], context)
        else:
            view = cache.fields_view_get(model, False, 'tree', context)

        tree = tree_view.ViewTree(view, model, res_id, domain=domain, context=context, action="/tree/action")
        if tree.toolbar:
            for tool in tree.toolbar:
                if tool.get('icon'):
                    tool['icon'] = icons.get_icon(tool['icon'])
                else:
                    tool['icon'] = False

        return dict(tree=tree, model=model)

    @expose()
    def default(self, id, model, view_id, domain, context):
        params = TinyDict()

        try:
            view_id = int(view_id)
        except:
            view_id = False

        params.ids = id
        params.view_ids = [view_id]
        params.model = model
        params.domain = domain
        params.context = context or {}

        return self.create(params)

    def sort_callback(self, item1, item2, field, sort_order="asc"):
        a = item1[field]
        b = item2[field]

        if(sort_order == "dsc"):
            return -cmp(a, b)

        return cmp(a, b)

    @expose('json')
    def data(self, ids, model, fields, field_parent=None, icon_name=None, domain=[], context={}, sort_by=None, sort_order="asc"):

        ids = ids or []

        if isinstance(ids, basestring):
            ids = [int(id) for id in ids.split(',')]

        if isinstance(fields, basestring):
            fields = eval(fields)

        if isinstance(domain, basestring):
            domain = eval(domain)

        if isinstance(context, basestring):
            context = eval(context)

        if field_parent and field_parent not in fields:
            fields.append(field_parent)

        proxy = rpc.RPCProxy(model)

        ctx = context or {}
        ctx.update(rpc.session.context.copy())

        if icon_name:
            fields.append(icon_name)

        fields_info = cache.fields_get(model, fields, ctx)
        result = proxy.read(ids, fields, ctx)

        if sort_by:
            result.sort(lambda a,b: self.sort_callback(a, b, sort_by, sort_order))

        # formate the data
        for field in fields:

            if fields_info[field]['type'] in ('float', 'integer'):
                for x in result:
                    if x[field]:
                        x[field] = '%s'%(x[field])

            if fields_info[field]['type'] in ('date',):
                for x in result:
                    if x[field]:
                        date = time.strptime(x[field], DT_FORMAT)
                        x[field] = time.strftime('%x', date)

            if fields_info[field]['type'] in ('datetime',):
                for x in result:
                    if x[field]:
                        date = time.strptime(x[field], DHM_FORMAT)
                        x[field] = time.strftime('%x %H:%M:%S', date)

            if fields_info[field]['type'] in ('one2one', 'many2one'):
                for x in result:
                    if x[field]:
                        x[field] = x[field][1]

            if fields_info[field]['type'] in ('selection'):
                for x in result:
                    if x[field]:
                        x[field] = dict(fields_info[field]['selection']).get(x[field], '')

        records = []
        for item in result:

            # empty string instead of bool and None
            for k, v in item.items():
                if v==None or (v==False and type(v)==bool):
                    item[k] = ''

            record = {}

            record['id'] = item.pop('id')
            record['action'] = url('/tree/open', model=model, id=record['id'])
            record['target'] = None

            record['icon'] = None

            if icon_name and item.get(icon_name):
                icon = item.pop(icon_name)
                record['icon'] = icons.get_icon(icon)

                if icon == 'STOCK_OPEN':
                    record['action'] = None
                    record['target'] = None

            record['children'] = []

            if field_parent and field_parent in item:
                record['children'] = item.pop(field_parent) or None

            record['items'] = item

            records += [record]

        return dict(records=records)

    def do_action(self, name, adds={}, datas={}):
        params, data = TinyDict.split(datas)

        model = params.model
        context = params._terp_context or {}
        ids = data.get('ids') or []

        ctx = rpc.session.context.copy()
        ctx.update(context)

        if ids:
            ids = [int(id) for id in ids.split(',')]

        id = (ids or False) and ids[0]

        if len(ids):
            from openerp.controllers import actions
            return actions.execute_by_keyword(name, adds=adds, model=model, id=id, ids=ids, context=ctx, report_type='pdf')
        else:
            raise common.message(_("No record selected!"))

    @expose()
    def report(self, **kw):
        return self.do_action('client_print_multi', datas=kw)

    @expose()
    def action(self, **kw):
        params, data = TinyDict.split(kw)

        action = params.data

        if not action:
            return self.do_action('tree_but_action', datas=kw)

        from openerp.controllers import actions

        ids = params.selection or []
        id = (ids or False) and ids[0]

        return actions.execute(action, model=params.model, id=id, ids=ids, report_type='pdf')

    @expose()
    def switch(self, **kw):

        params, data = TinyDict.split(kw)

        ids = params.selection or []            
        if len(ids):
            from openerp.controllers import actions
            return actions.execute_window(False, res_id=ids, model=params.model, domain=params.domain)
        else:
            raise common.message(_('No resource selected!'))

    @expose()
    def open(self, **kw):
        datas = {}

        datas['_terp_model'] = kw.get('model')
        datas['_terp_context'] = kw.get('context', {})
        datas['_terp_domain'] = kw.get('domain', [])

        datas['ids'] = kw.get('id')

        return self.do_action('tree_but_open', datas=datas)

# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following 
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be 
#     kept as in original distribution without any changes in all software 
#     screens, especially in start-up page and the software header, even if 
#     the application source code has been changed or updated or code has been 
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
# 
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

import time
import socket
import xmlrpclib

import tiny_socket

from openerp import common


class RPCException(Exception):
    """A common exeption class for RPC errors.
    """

    def __init__(self, code, backtrace):

        self.code = code
        self.args = backtrace

        if hasattr(code, 'split'):
            lines = code.split('\n')

            self.type = lines[0].split(' -- ')[0]
            self.message = ''
            if len(lines[0].split(' -- ')) > 1:
                self.message = lines[0].split(' -- ')[1]

            self.data = '\n'.join(lines[2:])

        else:
            self.type = 'error'
            self.message = backtrace
            self.data = backtrace

        self.backtrace = backtrace

    def __str__(self):
        return self.message


class RPCGateway(object):
    """Gateway abstraction, that implement common stuffs for rpc gateways.
    All RPC gateway should extends this class.
    """

    def __init__(self, session):
        if not isinstance(session, RPCSession):
            raise TypeError("RPCSession argument expected, got %s" % type(session))
        self.session = session

    def __rpc__(self, obj, method, args=(), auth=True):
        """Derived classes should owverride this method.

        @param obj: the remote object
        @param method: the method of the remote object
        @param args: arguments to be passed
        @param oauth: authentication is required or not

        @return: the result of the method
        """
        pass

    @property
    def connection_string(self):
        """Get the connection string...
        """
        return "%s://%s:%s/"%(self.session.protocol, self.session.host, self.session.port)

    def __convert(self, result):

        if isinstance(result, str):
            # try to convert into unicode string
            try:
                return ustr(result)
            except Exception, e:
                return result

        elif isinstance(result, list):
            return [self.__convert(val) for val in result]

        elif isinstance(result, tuple):
            return tuple([self.__convert(val) for val in result])

        elif isinstance(result, dict):
            newres = {}
            for key, val in result.items():
                newres[key] = self.__convert(val)

            return newres

        else:
            return result

    def __execute(self, obj, method, args=(), auth=True):
        try:
            
            #print "TERP-CALLING:", obj, method, args
            result = self.__rpc__(obj, method, args, auth=auth)
            #print "TERP-RESULT:", result
            return self.__convert(result)

        except socket.error, (e1, e2):
            raise common.message(_('Connection refused!'))

        except RPCException, err:
            if err.type in ('warning', 'UserError'):
                if err.message in ('ConcurrencyException') and len(args) > 4:
                    raise common.concurrency(err.message, err.data, args)
                else:
                    raise common.warning(err.data)
            else:
                raise common.error(_('Application Error!'), err.backtrace)

        except Exception, e:
            raise common.error(_('Application Error!'), str(e))

    def execute(self, obj, method, *args):
        """Excecute the method of the obj with the given arguments.

        @param obj: the remote object
        @param method: the method of the remote object
        @param args: arguments to be passed

        @return: the result of the method
        """
        return self.__execute(obj, method, args)

    def execute_noauth(self, obj, method, *args):
        """Excecute the method of the obj with the given arguments without authentication.

        @param obj: the object
        @param method: the method to execute
        @param args: the arguments

        @return: the result of the method
        """
        return self.__execute(obj, method, args, auth=False)


class XMLRPCGateway(RPCGateway):
    """XMLRPC implementation.
    """

    def __init__(self, session):
        """Create new instance of XMLRPCGateway.

        @param session: a session
        """
        super(XMLRPCGateway, self).__init__(session)
        self._url = self.connection_string + 'xmlrpc/'

    def __rpc__(self, obj, method, args=(), auth=True):
        sock = xmlrpclib.ServerProxy(self._url + str(obj))
        try:
            if auth:
                args = (self.session.db, self.session.uid, self.session.password) + args
            return getattr(sock, method)(*args)
        except xmlrpclib.Fault, err:
            raise RPCException(err.faultCode, err.faultString)


class NETRPCGateway(RPCGateway):
    """NETRPC Implementation.
    """

    def __rpc__(self, obj, method, args=(), auth=True):
        sock = tiny_socket.mysocket()
        try:
            sock.connect(self.session.host, self.session.port)
            if auth:
                args = (self.session.db, self.session.uid, self.session.password) + args
            sock.mysend((obj, method) + args)
            res = sock.myreceive()
            sock.disconnect()
            return res

        except xmlrpclib.Fault, err:
            raise RPCException(err.faultCode, err.faultString)
        
        except tiny_socket.Myexception, err:
            raise RPCException(err.faultCode, err.faultString)


# XXX: Fix openerp server to return PyTZ compatible timezone name
_TZ_ALIASES = {
    'IST' : 'Asia/Calcutta'
}


class RPCSession(object):
    """Maintains client session and provides way to authenticate 
    client & invoce RPC requested by clients.
    """

    __slots__ = ['host', 'port', 'protocol', 'storage', 'gateway']

    def __init__(self, host, port, protocol='socket', storage={}):
        """Create new instance of RPCSession.

        @param host: the openerp-server host
        @params port: the openerp-server port
        @params protocol: the openerp-server protocol
        @param storage: a dict like storage that will be used to store session data
        """
        self.host = host
        self.port = port
        self.protocol = protocol
        self.storage = storage

        if protocol in ('http', 'https'):
            self.gateway = XMLRPCGateway(self)

        elif protocol == 'socket':
            self.gateway = NETRPCGateway(self)

        else:
            raise common.message(_("Unsupported protocol."))

    def __getattr__(self, name):
        try:
            return super(RPCSession, self).__getattribute__(name)
        except:
            pass

        return self.storage.get(name)

    def __setattr__(self, name, value):
        if name in self.__slots__:
            super(RPCSession, self).__setattr__(name, value)
        else:
            self.storage[name] = value

    def __getitem__(self, name):
        return self.storage[name]

    def __setitem__(self, name, value):
        self.storage[name] = value

    def __delitem__(self, name):
        try:
            del self.storage[name]
        except:
            pass

    def get(self, name, default=None):
        return self.storage.get(name, default)

    @property
    def context(self):
        return (self._context or {}).copy()

    @property
    def connection_string(self):
        return self.gateway.connection_string

    def listdb(self):
        try:
            return self.execute_noauth('db', 'list')
        except Exception, e:
            return -1

    def login(self, db, user, password):
        
        if not (db and user and password):
            return -1

        try:
            uid = self.execute_noauth('common', 'login', db, user, password)
        except Exception, e:
            return -1

        if uid <= 0:
            return -1

        self.uid = uid
        self.db = db
        self.password = password
        self.open = True

        # read the full name of the user
        self.user_name = self.execute('object', 'execute', 'res.users', 'read', [uid], ['name'])[0]['name']

        # set the context
        self.context_reload()

        return uid

    def logout(self):
        try:
            self.storage.clear()
        except Exception, e:
            pass

    def is_logged(self):
        return self.uid and self.open

    def context_reload(self):
        """Reload the context for the current user
        """

        self._context = {'client': 'web'}
        self.timezone = 'utc'

        # self.uid
        context = self.execute('object', 'execute', 'res.users', 'context_get')
        self._context.update(context or {})
        
        if self.context.get('tz', False):
            self.timezone = self.execute('common', 'timezone_get')
            try:
                import pytz
            except:
                raise common.warning(_('You select a timezone but OpenERP could not find pytz library!\nThe timezone functionality will be disable.'))
                
        # set locale in session
        self.locale = self.context.get('lang')

    def execute(self, obj, method, *args):
        if not self.is_logged():
            raise common.warning(_('Not logged...'), _('Authorization Error!'))

        return self.gateway.execute(obj, method, *args)

    def execute_noauth(self, obj, method, *args):
        return self.gateway.execute_noauth(obj, method, *args)
    
    def execute_db(self, method, *args):
        return self.execute_noauth('db', method, *args)


# global session variable, will be initialized with connect
session = None


def initialize(host, port, protocol='socket', storage=None):
    """ Initialize the default rpc session.
    """
    global session
    session = RPCSession(host, port, protocol, storage=storage)


class RPCProxy(object):
    """A wrapper arround xmlrpclib, provides pythonic way to access tiny resources.

    For example,

    >>> users = RPCProxy("ir.users")
    >>> res = users.read([1], ['name', 'active_id'], session.context)
    """

    def __init__(self, resource):
        """Create new instance of RPCProxy for the give tiny resource

        @param resource: the tinyresource
        """
        self._resource = resource
        self._session = session
        self._attrs = {}
        
    def _func_getter(self, name):
        return lambda *args: self._session.execute("object", "execute", self._resource, name, *args)

    def __getattr__(self, name):
        if name not in self._attrs:
            return self._attrs.setdefault(name, self._func_getter(name))
        return self._attrs[name]



if __name__=="__main__":

    host = 'localhost'
    port = 8070
    protocol = 'socket'

    initialize(host, port, protocol, storage=dict())

    res = session.listdb()
    print res

    res = session.login('t1', 'admin', 'admin')
    print res

    res = RPCProxy('res.users').read([session.uid], ['name'])
    print res

    print session.context


# vim: ts=4 sts=4 sw=4 si et




import re
import collections

from openerp import tools


class Enum(set):

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError


class OrderedSet(set):

    def __init__(self, iterable=[]):

        self._items = items = []

        for item in iterable:
            self.add(item)

    def __contains__(self, value):
        return value in self._items

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)

    def add(self, value):
        if value not in self._items:
            self._items.append(value)

    def add_all(self, values):
        [self.add(value) for value in values]

    def discard(self, value):
        if value in self._items:
            self._items.remove(value)


class Bunch(dict):
    __setattr__ = dict.__setitem__

    def __delattr__(self, name):
        try:
            del self[name]
        except KeyError:
            raise AttributeError(name)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)


def make_bunch(d):
    """Converts a dict instance into a Bunch"""
    return Bunch(d)





###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

import os
import re

from openerp.tools import expose
from openerp.tools import redirect
from openerp.tools import find_resource
from openerp.tools import url

import cherrypy

from openerp import rpc
from openerp import common

from openerp import controllers
from openerp import cache

from openerp.controllers.base import SecuredController, unsecured
from openerp.controllers.base import login as tiny_login


def _cp_on_error():
    errorpage = cherrypy.request.app.root.errorpage
    message = errorpage.render()
    cherrypy.response.status = 500
    #cherrypy.response.headers['Content-Type'] = 'text/html'
    cherrypy.response.body = [message]

class Root(SecuredController):

    _cp_config = {'request.error_response': _cp_on_error}


    def user_action(self, id='action_id'):
        """Perform default user action.

        @param id: `action_id` or `menu_id`
        """

        proxy = rpc.RPCProxy("res.users")
        act_id = proxy.read([rpc.session.uid], [id, 'name'], rpc.session.context)

        if not act_id[0][id]:
            common.warning(_('You can not log into the system!\nAsk the administrator to verify\nyou have an action defined for your user.'), _('Access Denied!'))
            rpc.session.logout()
            raise redirect('/');

        act_id = act_id[0][id][0]

        return controllers.actions.execute_by_id(act_id)

    @expose()
    def index(self):
        """Index page, loads the view defined by `action_id`.
        """
        return self.user_action('action_id')

    @expose()
    def menu(self):
        """Main menu page, loads the view defined by `menu_id`.
        """
        return self.user_action('menu_id')

    @expose(allow_json=True)
    @unsecured
    def login(self, db=None, user=None, password=None, style=None, location=None, **kw):

        location = url(location or '/', kw or {})

        if db and user and user.startswith("anonymous"):
            if rpc.session.login(db, user, password):
                raise redirect(location)

        if cherrypy.request.params.get('tg_format') == 'json':
            if rpc.session.login(db, user, password) > 0:
                return dict(result=1)
            return dict(result=0)

        if style in ('ajax', 'ajax_small'):
            return dict(db=db, user=user, password=password, location=location, 
                    style=style, cp_template="templates/login_ajax.mako")

        return tiny_login(target=location, db=db, user=user, password=password, action="login")

    @expose()
    @unsecured
    def get_logo(self):
        
        comp_url = cherrypy.request.app.config['openerp-web'].get('company.url', None)

        res="""<img src="/static/images/openerp_big.png" alt="%(alt)s" border="0" width="200px" height="60px" usemap="#logo_map"/>
                    <map name="logo_map">
                        <area shape="rect" coords="102,42,124,56" href="http://openerp.com" target="_blank"/>
                        <area shape="rect" coords="145,42,184,56" href="http://axelor.com" target="_blank"/>
                    </map>"""%(dict(alt=_('OpenERP')))

        if os.path.exists(find_resource("openerp", "static/images/company_logo.png")):
            if comp_url:
                res = """   <a href='"""+comp_url+"""' target='_blank'>
                                <img src='/static/images/company_logo.png' alt="" border="0" width="205px" height="58px"/>
                            </a> """
            else:
                 res = """<img src="/static/images/company_logo.png" alt="" border="0" width="205px" height="58px"/>"""
        return res

    @expose()
    @unsecured
    def developped_by(self):
        comp_url = cherrypy.request.app.config['openerp-web'].get('company.url', None)

        res="""<img src="/static/images/developped_by.png" border="0" width="200" height="60" alt="%(alt)s" usemap="#devby_map"/>
                    <map name="devby_map">
                        <area shape="rect" coords="0,20,100,60" href="http://axelor.com" target="_blank"/>
                        <area shape="rect" coords="120,20,200,60" href="http://openerp.com" target="_blank"/>
                    </map>"""%(dict(alt=_('Developped by Axelor and Tiny')))

        if os.path.exists(find_resource("openerp", "static/images/company_logo.png")):
            if comp_url:
                res = """   <a href='"""+comp_url+"""' target='_blank'>
                                <img src='/static/images/company_logo.png' alt="" border="0" width="205px" height="58px"/>
                            </a> """
            else:
                 res = """<img src="/static/images/company_logo.png" alt="" border="0" width="205px" height="58px"/>"""
        return res

    @expose()
    @unsecured
    def logout(self):
        """ Logout method, will terminate the current session.
        """
        rpc.session.logout()
        raise redirect('/')

    @expose(template="templates/about.mako")
    @unsecured
    def about(self):
        from openerp import release
        version = _("Version %s") % (release.version,)
        return dict(version=version)
    
    profile = profile.profiler

    form = controllers.form.Form()
    tree = controllers.tree.Tree()
    wizard = controllers.wizard.Wizard()
    search = controllers.search.Search()
    pref = controllers.preferences.Preferences()
    selection = controllers.selection.Selection()
    shortcuts = controllers.shortcuts.Shortcuts()
    requests = controllers.requests.Requests()
    openm2o = controllers.openm2o.OpenM2O()
    openo2m = controllers.openo2m.OpenO2M()
    openm2m = controllers.openm2m.OpenM2M()
    listgrid = controllers.listgrid.List()
    attachment = controllers.attachment.Attachment()
    translator = controllers.translator.Translator()
    impex = controllers.impex.ImpEx()
    fieldpref = controllers.fieldpref.FieldPref()
    calendar = controllers.tinycalendar.TinyCalendar()
    calpopup = controllers.tinycalendar.CalendarPopup()
    viewlog = controllers.view_log.View_Log()
    image = controllers.image.Image()
    database = controllers.database.Database()
    viewed = controllers.viewed.ViewEd()
    viewlist = controllers.viewlist.ViewList()
    workflow = controllers.workflow.Workflow()
    workflowlist = controllers.workflow.WorkflowList()
    process = controllers.process.Process()
    wiki = controllers.wiki.WikiView()
    errorpage = controllers.error_page.ErrorPage()


# vim: ts=4 sts=4 sw=4 si et




import copy

from itertools import ifilter, count, chain, izip, ifilterfalse

import cherrypy

from openerp import tools

from openerp.validators import *

from meta import WidgetType
from utils import OrderedSet
from utils import make_bunch


__all__ = ['Widget', 'InputWidget']


_serial_generator = count()

class Widget(object):
    
    __metaclass__ = WidgetType

    template = None

    css = []
    javascript = []

    css_class = None
    css_classes = []

    params = {
        'name': 'The Name for this Widget.',
        'css_class': 'Main CSS class for this widget',
        'css_classes': 'List of all CSS classes',
    }
    
    member_widgets = []
    default = None
    parent = None
    
    def __init__(self, name=None, **params):
        # set each keyword args as attribute
        for k, v in params.iteritems():
            if not k.startswith('_'):
                try:
                    setattr(self, k, v)
                except AttributeError, e:
                    #skip setting the value of a read only property
                    pass

        self._name = name
        self._serial = _serial_generator.next()
        
        # params & member_widgets
        for name in chain(self.__class__.params, self.__class__.member_widgets):
            try:
                attr = getattr(self, name, None)
                if isinstance(attr, list):
                    attr = copy.copy(attr)
                elif isinstance(attr, dict):
                    attr = attr.copy()
                setattr(self, name, attr)
            except AttributeError, e:
                pass

        self._resources = OrderedSet()

        # Set default css class for the widget
        if not getattr(self, 'css_class', None):
            self.css_class = self.__class__.__name__.lower()
            
    @property
    def name(self):
        return self._name
    
    @property
    def path(self):
        item = self
        while item:
            yield item
            item = item.parent
            
    @property
    def root(self):
        return list(self.path)[-1]
    
    @property
    def is_root(self):
        return self.parent is None
    
    def adjust_value(self, value, **params):
        """Adjust the value sent to the template on display."""
        if value is None:
            value = self.default
            if callable(value):
                value = value()
        return value
    
    def iter_member_widgets(self):
        """Iterates over all the widget's children
        """
        for member in self.__class__.member_widgets:
            attr = getattr(self, member, None)
            if isinstance(attr, list):
                for widget in attr:
                    yield widget
            elif isinstance(attr, dict):
                for name, widget in attr.iteritems():
                    yield widget
            elif attr is not None:
                yield attr
    
    def value_for(self, item, value):
        """
        Get value for member widget.

        Pick up the value for a given member_widget 'item' from the
        value dict passed to this widget.
        """
        
        if getattr(item, 'strip_name', False):
            return value
        
        name = getattr(item, "name", item)
        if isinstance(value, dict):
            return value.get(name)
        else:
            return None
    
    def params_for(self, item, **params):
        """
        Get params for member widget.

        Pick up the params for the given member_widget 'item' from
        the params dict passed to this widget.
        """
        
        if getattr(item, 'strip_name', False):
            return params
        
        name = getattr(item, "name", item)
        item_params = {}
        for k, v in params.iteritems():
            if isinstance(v, dict):
                if name in v:
                    item_params[k] = v[name]
        return item_params
    
    def display_member(self, item, value, **params):
        
        if isinstance(item, basestring):
            item = getattr(self, item, None)
        
        assert isinstance(item, Widget), "Invalid member widget."
        
        v = self.value_for(item, value)
        d = self.params_for(item, **params)
        return item.display(v, **d)
    
    def update_params(self, params):
        
        # Populate dict with attrs from self listed at params and member_widgets
        for k in ifilterfalse(params.__contains__, chain(self.params, self.member_widgets)):
            attr = getattr(self, k, None)
            if attr is not None:
                if not isinstance(attr, Widget) and callable(attr):
                    attr = attr()
            params[k] = attr
            
        for w in self.iter_member_widgets():
            w.parent = self
            
        v = params['value']
        d = params['member_widgets_params']
        
        params['value_for'] = lambda f: self.value_for(f, v)
        params['params_for'] = lambda f: self.params_for(f, **d)
        params['display_member'] = lambda f: self.display_member(f, v, **d)
        
        params.css_class = ' '.join(set([params['css_class'] or ''] + params['css_classes']))
        
    def display(self, value=None, **params):
        
        params['member_widgets_params'] = params.copy()
        
        d = make_bunch(params)
        d.value = self.adjust_value(value, **params)
        
        self.update_params(d)
        
        return tools.render_template(self.template_c, d)
    
    def render(self, value=None, **params):
        return self.display(value, **params)
    
    def retrieve_css(self):
        """
        Return the needed CSS ressources.

        Return a setlike instance with all the CSSLinks and CSSSources
        the widget needs.
        """
        css = OrderedSet()
        for cssitem in self.css:
            css.add(cssitem)
        for widget in self.iter_member_widgets():
            for cssitem in widget.retrieve_css():
                css.add(cssitem)
        return css
    
    def retrieve_javascript(self):
        """
        Get JavaScript for the member widgets.

        Retrieve the JavaScript for all the member widgets and
        get an ordered union of them.
        """
        scripts = OrderedSet()
        for script in self.javascript:
            scripts.add(script)
        for widget in self.iter_member_widgets():
            for script in widget.retrieve_javascript():
                scripts.add(script)
        return scripts
       
    def __ne__(self, other):
        return not (self == other)

    def __eq__(self, other):
        return (
            (getattr(other, '__class__', None) is self.__class__) and
              (other._serial == self._serial) and
              (other._name == self._name)
            )
    
    def __repr__(self):
        return self.__class__.__name__
        return "%s(%s)" % (self.__class__.__name__, ', '.join(
            ["%s=%r" % (var, getattr(self, var))
                for var in ['name'] + list(self.__class__.params)]))


class InputWidget(Widget):
    
    params = {
        'required': "Whether the field value is required.",
        'readonly': "Whether the field is readonly.",
        'disabled': "Whether the field is disabled.",
    }
    
    validator = DefaultValidator

    required = False
    readonly = False
    disabled = False
    
    strip_name = False
    
    def __init__(self, name=None, **params):
        super(InputWidget, self).__init__(name, **params)
        
    @property
    def full_name(self):
        return ".".join(reversed([w.name for w in self.path if w.name and not getattr(w, 'strip_name', False)]))
    
    @property
    def is_validated(self):
        try:
            return self.root is cherrypy.request.validated_form
        except:
            pass
        return False

    @property
    def is_required(self):
        if self.required:
            return True
        try:
            self.validator.to_python('', None)
        except:
            return True
        return False

    @property
    def is_disabled(self):
        return self.disabled

    @property
    def is_readonly(self):
        return self.readonly

    def validate(self, value, state=None):
        """Validate value using validator if widget has one. If validation fails
        a formencode.Invalid exception will be raised.
        """
        if self.validator:
            try:
                value = self.validator.to_python(value, state)
            except Invalid, error:
                raise
        return value
    
    def safe_validate(self, value): 
        """Tries to coerce the value to python using the validator. If
        validation fails the original value will be returned unmodified."""
        
        # don't validate empty values
        if value is None or (isinstance(value, basestring) and value.strip() == ""):
            return value
        
        try:
            value = self.validate(value)
        except Exception:
            pass
        return value
    
    def adjust_value(self, value, **params):
        """Adjusts the python value sent to :meth:`Widget.display` with
        the validator so it can be rendered in the template.
        """
        
        iv = None        
        if hasattr(cherrypy.request, 'input_values') and self.is_validated:
            iv = cherrypy.request.input_values.get(self._name)
            
        if iv is not None and not isinstance(self.validator, (ForEach, Schema)):
            value = self.safe_validate(iv)
        else:
            value = super(InputWidget, self).adjust_value(value, **params)
        
        if self.validator and not isinstance(self.validator, Schema):
            # Does not adjust_value with Schema because it will recursively
            # call from_python on all sub-validators and that will send
            # strings through their update_params methods which expect
            # python values. adjust_value is called just before sending the
            # value to the template, not before.
            try:
                value = self.validator.from_python(value)
            except Exception, e:
                # Ignore conversion errors so bad-input is redisplayed properly
                pass
        return value
    
    def update_params(self, params):
        
        error = None
        if self.is_validated:
            error = getattr(cherrypy.request, 'validation_exception', None)
            value = getattr(cherrypy.request, 'validation_value', None)
                        
            if self.is_root:
                params['error'] = params.setdefault('error', error)
            elif error:
                params['error'] = params.setdefault('error', error.error_dict.get(self._name, None))
        else:
            params['error'] = params.setdefault('error', None)
            
        super(InputWidget, self).update_params(params)
            
        if not self.strip_name:
            
            if self.is_required:
                params.css_classes.append('requiredfield')
                
            if self.is_readonly:
                params.css_classes.append('readonlyfield')
            
            if self.is_disabled:
                params.css_classes.append('disabledfield')
                
            if params.error:
                params.css_classes.append('errorfield')

        params['error_for'] = lambda f: self.error_for(f, params['error'])
        params['css_class'] = ' '.join(set([params['css_class'] or ''] + params['css_classes']))
        
    def error_for(self, item, error):
        
        if getattr(item, "strip_name", False):
            return error

        try:
            error.error_dict[item.name]
        except:
            pass
        return None





###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

import base64
import os

from openerp.tools import expose
from openerp.tools import redirect

import cherrypy

from openerp import rpc
from openerp.controllers.base import SecuredController

from openerp.widgets.binary import get_temp_file

class Image(SecuredController):

    @expose(template="templates/image.mako")
    def index(self, **kw):

        saved = kw.get('saved') or None
        model = kw.get('model')
        id = kw.get('id')
        field = kw.get('field')

        return dict(model=model, saved=saved, id=id, field=field)

    @expose(content_type='application/octet')
    def get_image(self, **kw):
        model = kw.get('model')
        field = kw.get('field')
        id = int(kw.get('id'))

        proxy = rpc.RPCProxy(model)
        res = proxy.read([id], [field])[0]
        res = res.get(field)

        if res:
            return base64.decodestring(res)
        else:
            return ''
        
    @expose(content_type='application/octet')
    def get_picture(self, **kw):
        fname = get_temp_file(**kw)
        data = open(fname).read()
        if not data:
            # empty file that have just been created by get_temp_file
            raise cherrypy.HTTPError(404)
        return data

    @expose(template="templates/image.mako")
    def add(self, upimage,  **kw):

        saved = kw.get('saved') or None

        datas = upimage.file.read()

        model = kw.get('model')
        id = int(kw.get('id'))
        field = kw.get('field')

        value = base64.encodestring(datas)

        data = {field: value}

        proxy = rpc.RPCProxy(model)
        res = proxy.write([id], data)

        if res:
            saved = 1

        return dict(model=model, saved=saved, id=id, field=field)

    @expose(template="templates/image.mako")
    def delete(self, **kw):

        saved = None
        model = kw.get('model')
        id = int(kw.get('id'))
        field = kw.get('field')

        proxy = rpc.RPCProxy(model)
        proxy.write([id], {field: False})

        return dict(model=model, saved=saved, id=id, field=field)

    @expose(content_type='application/octet')
    def save_as(self, **kw):

        model = kw.get('model')
        id = int(kw.get('id'))
        field = kw.get('field')

        proxy = rpc.RPCProxy(model)
        res = proxy.read([id], [field])[0]

        res = res.get(field)

        if not res:
            raise redirect('/image', **kw)

        return base64.decodestring(res)

    @expose()
    def b64(self, **kw):
        #idea from http://dean.edwards.name/weblog/2005/06/base64-ie/
        try:
            qs = cherrypy.request.query_string
            content_type, data = qs.split(';')
            data_type, data = data.split(',')
            assert(data_type == 'base64')
            cherrypy.response.headers['Content-Type'] = content_type
            return base64.decodestring(data)
        except:
            raise cherrypy.HTTPError(400)   # Bad request

# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

from openerp.tools import expose
from openerp.tools import redirect

import cherrypy

from openerp import rpc
from openerp import tools
from openerp import common

from openerp.controllers.base import SecuredController

class Requests(SecuredController):

    def my(self):

        if not rpc.session.is_logged():
            return [],[]

        ids, ids2 = cherrypy.session.get('terp_requests', (False, False))
        if ids == False:
            ids, ids2 = rpc.RPCProxy('res.request').request_get()
            cherrypy.session['terp_requests'] = (ids, ids2)

        msg = _("No request")
        if len(ids):
            msg = _('%s request(s)') % len(ids)

        if len(ids2):
            msg += _(' - %s pending request(s)') % len(ids2)

        return ids, msg

    @expose()
    def default(self, ids):
        from openerp.controllers import actions
        #ids = eval(ids)

        #read requests
        ids, ids2 = rpc.RPCProxy('res.request').request_get()

        return actions.execute_window(False, 'res.request', res_id=None, domain=[('act_to','=',rpc.session.uid)], view_type='form', mode='tree,form')

# vim: ts=4 sts=4 sw=4 si et




import os
import urllib

import cherrypy
from mako.filters import html_escape

from openerp.validators import Invalid


__all__ = ["url", "url_plus", "redirect", "validate", "error_handler", "exception_handler", 
           "attrs", "attr_if", "decorated"]


def url(_cppath, _cpparams=None, _cpquote=False, **kw):
    """
    Returns absolute url for the given _cppath, _cpparams and kw.
    
    If _cppath is a list, path will be created joining them with '/'.
    If _cpquote if True, url_quote the params/kw
    If _cpparams is given it should be a map or list of tuples to create a map.
    
    query string will be created from _cpparams and **kw.
    
    >>> url("/some/path", {"a": 1, "b": 2})
    >>> /some/path?a=1&b=2
    >>> url(["some", "path"], a=1, b=2)
    >>> /some/path?a=1&b=2
    """
    
    path = _cppath
    if not isinstance(_cppath, basestring):
        path = "/".join(list(_cppath))

    params = _cpparams or {}
    if isinstance(_cpparams, list):
        params = dict(_cpparams)
    params.update(kw)
    
    kv = []
    for k, v in params.iteritems():
        if _cpquote and isinstance(k, basestring) and isinstance(v, basestring):
            k = urllib.quote_plus(k)
            v = urllib.quote_plus(v)
        kv.append("%s=%s" % (k, v))

    query = '&'.join(kv)

    if path and query:
        path = path + '?' + query
    elif query:
        path = query

    if path.startswith('/'):
        webpath = (cherrypy.config.get('server.webpath') or '').rstrip('/')
        if cherrypy.request.app:
            path = cherrypy.request.app.script_name + path
        path = webpath + path

    return path


def url_plus(_cppath, _cpparams=None, **kw):
    return url(_cppath, _cpparams, True, **kw)


def redirect(_cppath, _cpparams=None, **kw):
    return cherrypy.HTTPRedirect(url(_cppath, _cpparams, **kw))


from itertools import izip, islice
from inspect import getargspec

def to_kw(func, args, kw):

    argnames, defaults = getargspec(func)[::3]
    defaults = defaults or []

    kv = zip(islice(argnames, 0, len(argnames) - len(defaults)), args)
    kw.update(kv)

    return args[len(argnames)-len(defaults):], kw

def from_kw(func, args, kw):

    argnames, defaults = getargspec(func)[::3]
    defaults = defaults or []

    newargs = [kw.pop(name) for name in islice(argnames, 0, len(argnames) - len(defaults)) if name in kw]
    newargs.extend(args)

    return newargs, kw


def validate(form=None, validators=None):

    def validate_wrapper(func):

        if callable(form) and not hasattr(form, "validate"):
            init_form = lambda self: form(self)
        else:
            init_form = lambda self: form

        def func_wrapper(*args, **kw):

            # do not validate a second time if already validated
            if hasattr(cherrypy.request, 'validation_state'):
                return func(*args, **kw)

            form = init_form(args and args[0] or kw["self"])
            args, kw = to_kw(func, args, kw)

            errors = {}

            if form:
                value = kw.copy()
                value.pop('self', None)
                try:
                    kw.update(form.validate(value, None))
                except Invalid, e:
                    errors = e.unpack_errors()
                    cherrypy.request.validation_exception = e
                    cherrypy.request.validation_value = value
                cherrypy.request.validated_form = form

            if validators:

                if isinstance(validators, dict):
                    for field, validator in validators.iteritems():
                        try:
                            kw[field] = validator.to_python(
                                kw.get(field, None), None)
                        except Invalid, error:
                            errors[field] = error
                else:
                    try:
                        value = kw.copy()
                        kw.update(validators.to_python(value, None))
                    except Invalid, e:
                        errors = e.unpack_errors()
                        cherrypy.request.validation_exception = e
                        cherrypy.request.validation_value = value

            cherrypy.request.validation_errors = errors
            cherrypy.request.input_values = kw.copy()
            cherrypy.request.validation_state = True

            args, kw = from_kw(func, args, kw)
            return func(*args, **kw)

        return decorated(func_wrapper, func)

    return validate_wrapper

def error_handler(handler):

    def wrapper(func):

        def func_wrapper(*args, **kw):

            tg_errors = getattr(cherrypy.request, 'validation_errors', None)
            if tg_errors:
                kw['tg_errors'] = tg_errors
                return handler(*args, **kw)

            return func(*args, **kw)

        return decorated(func_wrapper, func)

    return wrapper

def exception_handler(*args, **kw):
    return lambda f: f


def config(key, section, default=None):
    """A handy function to access config values.
    """
    return cherrypy.request.app.config.get(section, {}).get(key, default)


class NoEscape(object):
    """A special callable class to prevent appying `html_escape` filter 
    by the default `content` filter.
    """

    def __init__(self, value):
        self.value = value

    def __call__(self, *args, **kw):
        try:
            return unicode(self.value(*args, **kw))
        except:
            pass
        return unicode(self.value)
    
    def encode(self, encoding):
        return self().encode(encoding)
    
    def __unicode__(self):
        return self()
    
    def __str__(self):
        return self()


def content(value):
    """A Mako filter to return unicode string according to the given value.
    
    If value is None return empty string.
    If value is instance of NoEscape return unicode string.
    If value is not None nor instance of NoEscape return unicode string applying `html_escape` filter.
    """
    if value is None:
        return ""

    if isinstance(value, NoEscape):
        return ustr(value)

    return html_escape(ustr(value))


def attrs(*args, **kw):

    kv = {}

    for arg in args:
        if isinstance(arg, dict):
            kv.update(arg)
        else:
            raise TypeError

    kv.update(kw)

    alias = {
        'css_class': 'class',
    }

    result = []
    for name, value in kv.iteritems():
        if callable(value):
            value = value()
        if value is not None:
            name = alias.get(name, name)
            result.append('%s="%s"' % (name, content(value)))
    
    return NoEscape(" ".join(result))


def attr_if(name, expression):
    return NoEscape((expression or '') and '%s="%s"' % (name, content(name)))


def decorated(wrapper, func, **attrs):
    """Update decorated wrapper of the func with given attrs
    and make sure to keep original metadata.
    """

    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    wrapper.__dict__ = func.__dict__.copy()
    wrapper.__module__ = func.__module__

    for k, v in attrs.iteritems():
        try:
            setattr(wrapper, k, v)
        except:
            pass

    return wrapper




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

import cherrypy

from openerp import tools

from openerp.widgets.base import Widget
from openerp.widgets.base import InputWidget

__all__ = ['TinyWidget', 'TinyInputWidget', 'ConcurrencyInfo']


_attrs_boolean = {
    'select': False,
    'nolabel': False,
    'required': False,
    'readonly': False,
}

def _boolean_attr(attrs, name):

    if name not in _attrs_boolean:
        return attrs.get(name)

    val = attrs.get(name)
    if isinstance(val, basestring) and val.lower() in ('false', 'none', '0'):
        return False

    return (attrs.get(name) and True) or _attrs_boolean.get(name)


class TinyWidget(Widget):

    params = [
        'colspan',
        'rowspan',
        'string',
        'nolabel',
        'visible',
        'model',
    ]

    colspan = 1
    rowspan = 1
    string = None
    nolabel = False
    visible = True
    model = None

    def __init__(self, **attrs):

        super(TinyWidget, self).__init__(**attrs)

        prefix = attrs.get('prefix', '')
        self._name = prefix + (prefix and '/' or '') + attrs.get('name', '')

        self.colspan = int(self.colspan)
        self.rowspan = int(self.rowspan)
        self.nolabel = _boolean_attr(attrs, 'nolabel')

        self.visible = True

        try:
            visval = attrs.get('invisible', 'False')
            ctx = attrs.get('context', {})
            self.invisible = eval(visval, {'context': ctx})
        except:
            pass

        self.attributes = attrs.get('attrs', {})

    def get_widgets_by_name(self, name, kind=Widget, parent=None):

        result = []
        parent = parent or self

        for wid in parent.iter_member_widgets():

            if wid.name == name and isinstance(wid, kind):
                result.append(wid)

            if getattr(wid, 'member_widgets', False):
                result += self.get_widgets_by_name(name, kind=kind, parent=wid)

        return result

    @property
    def name(self):
        return self._name

    def get_last_update_info(resource, values):
        result = {}
        for item in values:
            result["%s,%s" % (resource, item['id'])] = item.pop('__last_update', '')
        return result

    def _update_concurrency_info(self, resource, records):
        info = getattr(cherrypy.request, 'terp_concurrency_info', {})
        vals = info.setdefault(resource, {})
        for item in records:
            vals[item['id']] = item.pop('__last_update', '')
        cherrypy.request.terp_concurrency_info = info


class TinyInputWidget(TinyWidget, InputWidget):

    params = [
        'select',
        'required',
        'readonly',
        'help',
        'editable',
        'translatable',
        'inline',
        'states',
        'callback',
        'change_default',
        'onchange',
        'kind',
    ]

    select = False
    required = False
    readonly = False
    help = None
    editable = True
    translatable = False
    inline = False

    states = None
    callback = None
    change_default = None
    onchange = 'onChange(this)'
    kind=None

    def __init__(self, **attrs):

        super(TinyInputWidget, self).__init__(**attrs)

        if isinstance(self.states, basestring):
            self.states = self.states.split(',')

        self.select = _boolean_attr(attrs, 'select')
        self.required = _boolean_attr(attrs, 'required')
        self.readonly = _boolean_attr(attrs, 'readonly')

        self.translatable = attrs.get('translate', False)

        self.set_state(attrs.get('state', 'draft'))

        self.callback = attrs.get('on_change', None)
        self.kind = attrs.get('type', None)

    def set_state(self, state):

        if isinstance(self.states, dict) and state in self.states:

            attrs = dict(self.states[state])

            if 'readonly' in attrs:
                self.readonly = attrs['readonly']

            if 'required' in attrs:
                self.required = attrs['required']

            if 'value' in attrs:
                self.default = attrs['value']

    def get_value(self):
        """Get the value of the field.

        @return: field value
        """
        return self.default

    def set_value(self, value):
        """Set the value of the field.

        @param value: the value
        """
        if isinstance(value, basestring):
            value = ustr(value)

        self.default = value

    def get_display_value(self):
        """Get the display value of the field.
        """

        try:
            return self.validator.from_python(self.default)
        except:
            pass

        return self.get_value()

    def update_params(self, d):
        super(TinyInputWidget, self).update_params(d)

        attrs = d['attrs'] = {}

        attrs['change_default'] = self.change_default or None
        attrs['callback'] = self.callback or None
        attrs['onchange'] = self.onchange

        d['kind'] = self.kind
        d['editable'] = self.editable
        d['inline'] = self.inline

        if self.readonly:
            attrs['disabled'] = 'disabled'

        if self.translatable and 'translatable' not in d['css_classes']:
            d.setdefault('css_classes', []).append("translatable")


class ConcurrencyInfo(TinyInputWidget):

    template="""
    % if ids and model in info:
        % for id in ids:
            % if id in info[model]:
                <input type="hidden" name="_terp_concurrency_info" value="('${model},${id}', '${info[model][id]}')"/>
            % endif
        % endfor
    % endif
    """

    params = ['ids', 'model', 'info']

    def __init__(self, model, ids):
        super(ConcurrencyInfo, self).__init__(model=model, ids=ids)

    @property
    def info(self):
        return getattr(cherrypy.request, 'terp_concurrency_info', {})


# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

from openerp.tools import expose
from openerp.tools import validate
from openerp.tools import error_handler
from openerp.tools import exception_handler

import cherrypy

from openerp import rpc
from openerp import cache
from openerp import tools
from openerp import widgets as tw

from openerp.utils import TinyDict

from form import Form
from form import get_validation_schema
from form import default_error_handler
from form import default_exception_handler

class OpenM2M(Form):

    path = '/openm2m'

    @expose(template="templates/openm2m.mako")
    def create(self, params, tg_errors=None):

        params.m2m = params.m2m or params.source
        params.editable = params.get('_terp_editable', True)
        params.hidden_fields = [tw.form.Hidden(name='_terp_m2m', default=params.m2m)]
        form = self.create_form(params, tg_errors)

        return dict(form=form, params=params)

    @expose()
    @validate(form=get_validation_schema)
    @error_handler(default_error_handler)
    @exception_handler(default_exception_handler)
    def save(self, terp_save_only=False, **kw):
        params, data = TinyDict.split(kw)

        # remember the current page (tab) of notebooks
        cherrypy.session['remember_notebooks'] = True

        # bypass save, for button action in non-editable view
        if not (params.button and not params.editable and params.id):

            proxy = rpc.RPCProxy(params.model)

            if not params.id:
                id = proxy.create(data, params.context)
                params.ids = (params.ids or []) + [int(id)]
                params.id = int(id)
                params.count += 1
            else:
                ctx = tools.context_with_concurrency_info(params.context, params.concurrency_info)
                id = proxy.write([params.id], data, ctx)

        current = params.chain_get(params.source or '')
        button = (params.button or False) and True
        
        params.load_counter = 1
        if current and current.id and not button:
            params.load_counter = 2

        # perform button action
        if params.button:
            res = self.button_action(params)
            if res:
                return res

        
        if current:
            current.id = None
            if not params.id:
                params.id = int(id)
        elif not button:
            params.editable = False

        if not current and not button:
            params.load_counter = 2

        return self.create(params)

    @expose()
    def new(self, **kw):
        params, data = TinyDict.split(kw)

        if not params.model:
            params.update(kw)

        params.view_mode = ['form', 'tree']
        params.view_type = 'form'

        params.editable = params.get('_terp_editable', True)

        return self.create(params)




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

import socket
import cPickle
import sys

DNS_CACHE = {}

class Myexception(Exception):
    def __init__(self, faultCode, faultString):
        self.faultCode = faultCode
        self.faultString = faultString
        self.args = (faultCode, faultString)

class mysocket:
    def __init__(self, sock=None):
        if sock is None:
            self.sock = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM)
        else:
            self.sock = sock
        self.sock.settimeout(120)

    def connect(self, host, port=False):
        if not port:
            protocol, buf = host.split('//')
            host, port = buf.split(':')
        if host in DNS_CACHE:
            host = DNS_CACHE[host]
        self.sock.connect((host, int(port)))
        DNS_CACHE[host], port = self.sock.getpeername()

    def disconnect(self):
        # on Mac, the connection is automatically shutdown when the server disconnect.
        # see http://bugs.python.org/issue4397
        if sys.platform != 'darwin':
            self.sock.shutdown(socket.SHUT_RDWR)
        self.sock.close()

    def mysend(self, msg, exception=False, traceback=None):
        msg = cPickle.dumps([msg,traceback])
        self.sock.sendall('%8d%s%s' % (len(msg), exception and "1" or "0", msg))

    def myreceive(self):
        def read(socket, size):
            buf=''
            while len(buf) < size:
                chunk = self.sock.recv(size - len(buf))
                if chunk == '':
                    raise RuntimeError, "socket connection broken"
                buf += chunk
            return buf

        size = int(read(self.sock, 8))
        buf = read(self.sock, 1)
        exception = buf != '0' and buf or False
        res = cPickle.loads(read(self.sock, size))

        if isinstance(res[0],Exception):
            if exception:
                raise Myexception(ustr(res[0]), ustr(res[1]))
            raise res[0]
        else:
            return res[0]

# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

"""This module implementes action methods.
"""

import time
import datetime
import base64

import cherrypy

from openerp import rpc
from openerp import tools
from openerp import common

from form import Form
from tree import Tree
from wizard import Wizard
from selection import Selection

from openerp.utils import TinyDict

def execute_window(view_ids, model, res_id=False, domain=None, view_type='form', context={},
                   mode='form,tree', name=None, target=None, limit=None):
    """Performs `actions.act_window` action.

    @param view_ids: view ids
    @param model: a model for which the action should be performed
    @param res_id: resource id
    @param domain: domain
    @param view_type: view type, eigther `form` or `tree`
    @param context: the context
    @param mode: view mode, eigther `form,tree` or `tree,form` or None

    @return: view (mostly XHTML code)
    """

    params = TinyDict()

    params.model = model
    params.ids = res_id
    params.view_ids = view_ids
    params.domain = domain or []
    params.context = context or {}
    params.limit = limit
    
    cherrypy.request._terp_view_name = name or None
    cherrypy.request._terp_view_target = target or None
    
    if params.ids and not isinstance(params.ids, list):
        params.ids = [params.ids]

    params.id = (params.ids or False) and params.ids[0]

    mode = mode or view_type

    if view_type == 'form':
        mode = mode.split(',')
        params.view_mode=mode

        return Form().create(params)

    elif view_type == 'tree':
        return Tree().create(params)

    else:
        raise common.message(_("Invalid View!"))

def execute_wizard(name, **datas):
    """Executes given wizard with the given data

    @param name: name of the wizard
    @param datas: datas

    @return: wizard view (mostly XHTML code)
    """
    params = TinyDict()
    params.name = name
    params.datas = datas
    params.state = 'init'

    return Wizard().create(params)

PRINT_FORMATS = {
     'pdf' : 'application/pdf',
     'doc' : 'application/vnd.ms-word',
     'html': 'text/html',
     'sxw' : 'application/vnd.sun.xml.writer'
}

def _print_data(data):

    if 'result' not in data:
        raise common.message(_('Error no report'))

    if data.get('code','normal')=='zlib':
        import zlib
        content = zlib.decompress(base64.decodestring(data['result']))
    else:
        content = base64.decodestring(data['result'])

    cherrypy.response.headers['Content-Type'] = PRINT_FORMATS[data['format']]
    return content

def execute_report(name, **data):
    """Executes a report with the given data, on success returns `application/pdf` data

    @param name: name of the report
    @param data: report data

    @return: `application/pdf` data
    """
    datas = data.copy()
    ids = datas['ids']
    del datas['ids']

    if not ids:
        ids =  rpc.session.execute('object', 'execute', datas['model'], 'search', [])
        if ids == []:
            raise common.message(_('Nothing to print!'))

        datas['id'] = ids[0]

    try:
        report_id = rpc.session.execute('report', 'report', name, ids, datas, rpc.session.context)
        state = False
        attempt = 0
        while not state:
            val = rpc.session.execute('report', 'report_get', report_id)
            state = val['state']
            if not state:
                time.sleep(1)
                attempt += 1
            if attempt>200:
                raise common.message(_('Printing aborted, too long delay!'))

        # report name
        report_name = 'report'
        report_type = val['format']

        if name != 'custom':
            proxy = rpc.RPCProxy('ir.actions.report.xml')
            res = proxy.search([('report_name','=', name)])
            if res:
                report_name = proxy.read(res[0], ['name'])['name']

        report_name = report_name.replace('Print ', '')
        cherrypy.response.headers['Content-Disposition'] = 'filename="' + report_name + '.' + report_type + '"';

        return _print_data(val)

    except rpc.RPCException, e:
        raise e

def execute(action, **data):
    """Execute the action with the provided data. for internal use only.

    @param action: the action
    @param data: the data

    @return: mostly XHTML code
    """

    if 'type' not in action:
        #XXX: in gtk client just returns to the caller
        #raise common.error('Error', 'Invalid action...')
        return

    if action['type'] == 'ir.actions.act_window_close':
        return """<html>
        <head>
            <script type="text/javascript">
                window.onload = function(evt){
                    if (window.opener) {
                        window.opener.setTimeout("window.location.reload()", 0);
                        window.close();
                    } else {
                        window.location.href = '/';
                    }
                }
            </script>
        </head>
        <body></body>
        </html>
        """

    elif action['type']=='ir.actions.act_window':
        for key in ('res_id', 'res_model', 'view_type','view_mode', 'limit'):
            data[key] = action.get(key, data.get(key, None))
            
        if not data.get('limit'):
            data['limit'] = 80

        view_ids=False
        if action.get('views', []):
            if isinstance(action['views'], list):
                view_ids=[x[0] for x in action['views']]
                data['view_mode']=",".join([x[1] for x in action['views']])
            else:
                if action.get('view_id', False):
                    view_ids=[action['view_id'][0]]
        elif action.get('view_id', False):
            view_ids=[action['view_id'][0]]

        if not action.get('domain', False):
            action['domain']='[]'

        ctx = data.get('context', {}).copy()
        ctx.update({'active_id': data.get('id', False), 'active_ids': data.get('ids', [])})
        ctx.update(tools.expr_eval(action.get('context', '{}'), ctx.copy()))
        
        # save active_id in session
        rpc.session.active_id = data.get('id')

        a = ctx.copy()
        a['time'] = time
        a['datetime'] = datetime
        domain = tools.expr_eval(action['domain'], a)

        if data.get('domain', False):
            domain.append(data['domain'])

        res = execute_window(view_ids,
                             data['res_model'],
                             data['res_id'],
                             domain,
                             action['view_type'],
                             ctx, data['view_mode'],
                             name=action.get('name'),
                             target=action.get('target'),
                             limit=data.get('limit'))
        
        return res

    elif action['type']=='ir.actions.server':

        ctx = data.get('context', {}).copy()
        ctx.update({'active_id': data.get('id', False), 'active_ids': data.get('ids', [])})

        res = rpc.RPCProxy('ir.actions.server').run([action['id']], ctx)
        if res:
            return execute(res, **data)

    elif action['type']=='ir.actions.wizard':
        if 'window' in data:
            del data['window']
        return execute_wizard(action['wiz_name'], **data)

    elif action['type']=='ir.actions.report.custom':
        data['report_id'] = action['report_id']
        return execute_report('custom', **data)

    elif action['type']=='ir.actions.report.xml':
        return execute_report(action['report_name'], **data)

    elif action['type']=="ir.actions.act_url":
        data['url'] = action['url']
        data['target'] = action['target']
        data['type'] = action['type']

        return execute_url(**data)

def execute_url(**data):
    url = data.get('url') or ''

    if not ('://' in url or url.startswith('/')):
        raise common.message(_('Relative URLs are not supported!'))

    raise tools.redirect(url)

def get_action_type(act_id):
    """Get the action type for the given action id.

    @param act_id: the action id
    @return: action type
    """
    res = rpc.session.execute('object', 'execute', 'ir.actions.actions', 'read', act_id, ['type'], rpc.session.context)

    if not (res and len(res)):
        raise common.message(_('Action not found!'))

    return res['type']

def execute_by_id(act_id, type=None, **data):
    """Perforns the given action of type `type` with the provided data.

    @param act_id: the action id
    @param type: action type
    @param data: the data

    @return: JSON object or XHTML code
    """

    if type==None:
        type = get_action_type(act_id)
	
    res = rpc.session.execute('object', 'execute', type, 'read', act_id, False, rpc.session.context)
    return execute(res, **data)

def execute_by_keyword(keyword, adds={}, **data):
    """Performs action represented by the given keyword argument with given data.

    @param keyword: action keyword
    @param data: action data

    @return: XHTML code
    """

    actions = None
    if 'id' in data:
        try:
            id = data.get('id', False)
            if (id != False): id = int(id)
            actions = rpc.session.execute('object', 'execute', 'ir.values', 'get', 'action', keyword, [(data['model'], id)], False, rpc.session.context)
            actions = map(lambda x: x[2], actions)
        except rpc.RPCException, e:
            raise e

    keyact = {}
    for action in actions:
        keyact[action['name']] = action

    keyact.update(adds)

    if not keyact:
        raise common.message(_('No action defined!'))

    if len(keyact) == 1:
        key = keyact.keys()[0]
        return execute(keyact[key], **data)
    else:
        return Selection().create(keyact, **data)

# vim: ts=4 sts=4 sw=4 si et




from openerp.tests.test import TestCase, setup_server

from openerp import rpc

class RootTest(TestCase):

    def test_index(self):
        self.getPage("/")
        self.assertStatus(200)

    def test_about(self):
        self.getPage("/about")
        self.assertStatus(200)

    def test_menu(self):
        self.getPage("/menu")
        self.assertStatus(200)
        self.assertInBody("new TreeGrid")
        




# -*- coding: UTF-8 -*-
"""This module contains functions called from console script entry points."""

import os
import sys
import optparse

from os.path import join, dirname, exists

import cherrypy
from cherrypy._cpconfig import as_dict
from formencode import NestedVariables

from openerp import release


__all__ = ['start']


def nestedvars_tool():
    if hasattr(cherrypy.request, 'params'):
        cherrypy.request.params = NestedVariables.to_python(cherrypy.request.params or {})

cherrypy.tools.nestedvars = cherrypy.Tool("before_handler", nestedvars_tool)
cherrypy.lowercase_api = True


class CPSessionWrapper(object):

    def __setattr__(self, name, value):
        cherrypy.session[name] = value

    def __getattr__(self, name):
        return cherrypy.session.get(name)

    def __delattr__(self, name):
        if name in cherrypy.session:
            del cherrypy.session[name]

    __getitem__ = __getattr__
    __setitem__ = __setattr__

    def get(self, name, default=None):
        return cherrypy.session.get(name, default)

    def clear(self):
        cherrypy.session.clear()


class ConfigurationError(Exception):
    pass


def get_config_file():
    setupdir = dirname(dirname(__file__))
    configfile = join(setupdir, "config", "openerp-web.cfg")
    if exists(configfile):
        return configfile
    return None


def setup_server(configfile):

    if not exists(configfile):
        raise ConfigurationError(_("Could not find configuration file: %s") % configfile)


    cherrypy.config.update({
        'tools.sessions.on':  True,
        'tools.nestedvars.on':  True
    })

    app_config = as_dict(configfile)
    
    _global = app_config.pop('global', {})
    _environ = _global.setdefault('server.environment', 'development')
    
    if _environ != 'development':
        _global['environment'] = _environ
        
    cherrypy.config.update(_global)
    
    static_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'static')
    app_config.update({'/static': {
        'tools.staticdir.on': True,
        'tools.staticdir.dir': static_dir
    }})
    
    app_config.update({'/favicon.ico': {
        'tools.staticfile.on': True,
        'tools.staticfile.filename': static_dir + "/images/favicon.ico"
    }})

    app_config.update({'/LICENSE.txt': {
        'tools.staticfile.on': True,
        'tools.staticfile.filename': static_dir + "/../../doc/LICENSE.txt"
    }})

    # import profiler while makes profile decorator available as __builtins__
    from openerp import profiler
    
    from openerp.controllers.root import Root
    app = cherrypy.tree.mount(Root(), '/', app_config)

    import pkg_resources
    from openerp.widgets import register_resource_directory

    static = pkg_resources.resource_filename("openerp", "static")
    register_resource_directory(app, "openerp", static)

    # initialize the rpc session
    host = app.config['openerp'].get('host')
    port = app.config['openerp'].get('port')
    protocol = app.config['openerp'].get('protocol')

    from openerp import rpc
    rpc.initialize(host, port, protocol, storage=CPSessionWrapper())


def start():
    """Start the CherryPy application server."""

    parser = optparse.OptionParser(version=release.version)
    parser.add_option("-c", "--config", dest="config", help="specify alternate config file", default=get_config_file())
    (opt, args) = parser.parse_args()

    setup_server(opt.config)

    cherrypy.engine.start()
    cherrypy.engine.block()



# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

from openerp import rpc
from openerp import tools

from openerp.utils import TinyDict

from screen import Screen

from base import JSSource
from interface import TinyWidget


class Sidebar(TinyWidget):

    template = "templates/sidebar.mako"
    params = ['reports', 'actions', 'relates', 'attachments']

    javascript = [JSSource("""
        function toggle_sidebar(forced) {

            var sb = MochiKit.DOM.getElement('sidebar');
            var sbp = MochiKit.DOM.getElement('sidebar_pane');

            sb.style.display = forced ? forced : (sb.style.display == "none" ? "" : "none");
            sbp.style.display = sb.style.display;

            set_cookie("terp_sidebar", sb.style.display);

            var img = MochiKit.DOM.getElementsByTagAndClassName('img', null, 'sidebar_hide')[0];
            if (sb.style.display == "none")
                img.src = '/static/images/sidebar_show.gif';
            else
                img.src = '/static/images/sidebar_hide.gif';
        }

        MochiKit.DOM.addLoadEvent(function(evt) {
            var sb = MochiKit.DOM.getElement('sidebar');
            if (sb) toggle_sidebar(get_cookie('terp_sidebar'));
        });
    """)]

    def __init__(self, model, toolbar=None, id=None, view_type="form", multi=True, is_tree=False, context={}, **kw):

        super(Sidebar, self).__init__(model=model, id=id)

        self.multi = multi
        self.context = context or {}
        self.view_type = view_type

        toolbar = toolbar or {}

        self.reports = toolbar.get('print', [])
        self.actions = toolbar.get('action', [])
        self.relates = toolbar.get('relate', [])

        self.attachments = []

        proxy = rpc.RPCProxy('ir.values')

        if not self.actions:

            act = 'client_action_multi'
            if is_tree:
                act = 'tree_but_action'

            res = []
            try: # Deal with `You try to bypass an access rule`
                res = proxy.get('action', act, [(self.model, False)], False, self.context)
            except:
                pass

            actions = [a[-1] for a in res]
            self.actions = [a for a in actions if self.multi or not a.get('multi')]

        if not self.reports:
            res = []
            try:
                res = proxy.get('action', 'client_print_multi', [(self.model, False)], False, self.context)
            except:
                pass

            actions = [a[-1] for a in res]
            self.reports = [a for a in actions if self.multi or not a.get('multi')]


        if self.view_type == 'form':
            
            proxy = rpc.RPCProxy('ir.attachment')
            ids = proxy.search([('res_model', '=', model), ('res_id', '=', id)], 0, 0, 0, self.context)
            
            if ids:
                attach = []
                datas = proxy.read(ids, ['datas_fname'])
                self.attachments = [(d['id'], d['datas_fname']) for d in datas if d['datas_fname']]
        else:
            self.relates = []

# vim: ts=4 sts=4 sw=4 si et




###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

import cherrypy

from openerp import tools
from openerp import rpc
from openerp import cache

from interface import TinyInputWidget

import form
import graph
import listgrid

import tinycalendar

class Screen(TinyInputWidget):

    template = """
        <input type="hidden" id="${name}_terp_model" name="${name}_terp_model" value="${model}"/>
        <input type="hidden" id="${name}_terp_state" name="${name}_terp_state" value="${state}"/>
        <input type="hidden" id="${name}_terp_id" name="${name}_terp_id" value="${id}"/>
        <input type="hidden" id="${name}_terp_ids" name="${name}_terp_ids" value="${ids}"/>

        <input type="hidden" id="${name}_terp_view_ids" name="${name}_terp_view_ids" value="${view_ids}"/>
        <input type="hidden" id="${name}_terp_view_mode" name="${name}_terp_view_mode" value="${view_mode}"/>
        <input type="hidden" id="${name}_terp_view_type" name="${name}_terp_view_type" value="${view_type}"/>
        <input type="hidden" id="${name}_terp_view_id" name="${name}_terp_view_id" value="${view_id}"/>
        <input type="hidden" id="${name}_terp_domain" name="${name}_terp_domain" value="${domain}"/>
        <input type="hidden" id="${name}_terp_context" name="${name}_terp_context" value="${ctx}"/>
        <input type="hidden" id="${name}_terp_editable" name="${name}_terp_editable" value="${editable}"/>

        <input type="hidden" id="${name}_terp_limit" name="${name}_terp_limit" value="${limit}"/>
        <input type="hidden" id="${name}_terp_offset" name="${name}_terp_offset" value="${offset}"/>
        <input type="hidden" id="${name}_terp_count" name="${name}_terp_count" value="${count}"/>

        % if widget:
            ${display_member(widget)}
        % endif
    """

    params = ['model', 'state', 'id', 'ids', 'view_id', 'view_ids', 'view_mode', 'view_type', 'domain', 'context', 'limit', 'offset', 'count']
    member_widgets = ['widget']

    def __init__(self, params=None, prefix='', name='', views_preloaded={}, hastoolbar=False, editable=False, readonly=False, selectable=0, nolinks=1):

        # get params dictionary
        params = params or cherrypy.request.terp_params
        prefix = prefix or (params.prefix or '')

        super(Screen, self).__init__(prefix=prefix, name=name)

        self.model         = params.model
        self.state         = params.state or None
        self.id            = params.id or False
        self.ids           = params.ids
        self.view_ids      = params.view_ids or []
        self.view_mode     = params.view_mode or []
        self.view_type     = params.view_type
        self.view_id       = False

        self.is_wizard = params.is_wizard

        while len(self.view_ids) < len(self.view_mode):
            self.view_ids += [False]

        if not self.view_type and params.view_mode:
            self.view_type = params.view_mode[0]

        if self.view_ids and self.view_type in self.view_mode:
            idx = self.view_mode.index(self.view_type)
            self.view_id = self.view_ids[idx]

        self.domain        = params.domain or []
        self.context       = params.context or {}
        self.nodefault     = params.nodefault or False

        self.offset        = params.offset
        self.limit         = params.limit
        self.count         = params.count

        if (self.ids or self.id) and self.count == 0:
            self.count = rpc.RPCProxy(self.model).search_count(self.domain)

        self.prefix             = prefix
        self.views_preloaded    = views_preloaded or (params.views or {})

        self.hastoolbar         = hastoolbar
        self.toolbar            = None

        self.selectable         = selectable
        self.editable           = editable
        self.readonly           = readonly
        self.link               = nolinks

        # get calendar options
        self.kalendar           = params.kalendar

        if self.view_mode:
            self.add_view_id(self.view_id, self.view_type)

    def add_view_id(self, view_id, view_type):
        self.view_id = view_id

        if view_type in self.views_preloaded:
            view = self.views_preloaded[view_type]
        else:
            ctx = rpc.session.context.copy()
            ctx.update(self.context)
            view = cache.fields_view_get(self.model, view_id, view_type, ctx, self.hastoolbar)

        self.add_view(view, view_type)

    def add_view(self, view, view_type='form'):

        self.view_id = view.get('view_id', self.view_id)

        if view_type == 'form':
            self.widget = form.Form(prefix=self.prefix,
                                    model=self.model,
                                    view=view,
                                    ids=(self.id or []) and [self.id],
                                    domain=self.domain,
                                    context=self.context,
                                    editable=self.editable,
                                    readonly=self.readonly,
                                    nodefault=self.nodefault, nolinks=self.link)

            if not self.is_wizard and self.ids is None:
                proxy = rpc.RPCProxy(self.model)
                self.ids = proxy.search(self.domain, self.offset or False, self.limit or False, 0, self.context)
                self.count = proxy.search_count(self.domain, self.context)

        elif view_type == 'tree':
            self.widget = listgrid.List(self.name or '_terp_list',
                                        model=self.model,
                                        view=view,
                                        ids=self.ids,
                                        domain=self.domain,
                                        context=self.context,
                                        editable=self.editable,
                                        selectable=self.selectable,
                                        offset=self.offset, limit=self.limit, count=self.count, nolinks=self.link)

            self.ids = self.widget.ids
            self.limit = self.widget.limit
            self.count = self.widget.count

        elif view_type == 'graph':
            self.widget = graph.Graph(model=self.model,
                                      view=view,
                                      view_id=view.get('view_id', False),
                                      ids=self.ids, domain=self.domain,
                                      context=self.context)
            self.ids = self.widget.ids

        elif view_type == 'calendar':
            self.widget = tinycalendar.get_calendar(view=view,
                                                    model=self.model,
                                                    ids=self.ids,
                                                    domain=self.domain,
                                                    context=self.context,
                                                    options=self.kalendar)

        elif view_type == 'gantt':
            self.widget = tinycalendar.GanttCalendar(model=self.model,
                                                     view=view,
                                                     ids=self.ids,
                                                     domain=self.domain,
                                                     context=self.context,
                                                     options=self.kalendar)

        self.string = (self.widget or '') and self.widget.string

        toolbar = {}
        for item, value in view.get('toolbar', {}).items():
            if value: toolbar[item] = value

        self.toolbar = toolbar or None
        self.hastoolbar = (toolbar or False) and True

# vim: ts=4 sts=4 sw=4 si et




import os
import time
import logging

from os import mkdir
from os.path import dirname, join, abspath, exists

from itertools import izip, islice
from inspect import getargspec

import cherrypy
from cherrypy.lib import profiler

prof_dir = abspath(cherrypy.config.get('server.profile_dir', 'profile'))
prof_on = cherrypy.config.get('server.profile_on', False)

if prof_on and not exists(prof_dir):
    mkdir(prof_dir)

__PROFILES = {}

def to_kw(func, args, kw):

    argnames, defaults = getargspec(func)[::3]
    defaults = defaults or []
    
    kw = kw.copy()

    kv = zip(islice(argnames, 0, len(argnames) - len(defaults)), args)
    kw.update(kv)

    return args[len(argnames)-len(defaults):], kw

_RPCTIME = 0

def profile(name, log=[], cb=None):
    """
    Profile decorator.
    
    @param name: name for the profile (should be unique)og
    @param log: list of argument name to be logged, for var args give 
                index value, will be logger in sequence
    @param cb: callback to be used to log custom message then from args
    
    >>> @profile("my_say", log=[0, 2, 'something', 'format'])
    >>> def say(what, something, tosay, format=None, *args):
    >>>     ...
    """
    
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    
    handler = logging.FileHandler(join(prof_dir, name + '.log'))
    logger.addHandler(handler)
    
    formatter = logging.Formatter("%(time)10.3f %(rpctime)10.3f %(efftime)10.3f -- %(message)s")
    handler.setFormatter(formatter)
    
    def message(fn, *args, **kw):
        
        if callable(cb):
            return "%s" % cb(*args, **kw)
        
        _a, kw = to_kw(fn, args, kw)
        
        res = []
        for n in log:
            if isinstance(n, int):
                try:
                    res.append("%s" % args[n])
                except:
                    pass
            else:
                res.append("%s=%s" % (n, kw.get(n)))
        return ", ".join(res)

    def wrapper(func):
        
        def func_wrapper(*args, **kw):
            
            global _RPCTIME
            
            rt = _RPCTIME
            
            t1 = time.time()
            res = func(*args, **kw)
            t2 = time.time()
            
            t = t2 - t1
            
            if name == "rpc.execute":
                _RPCTIME += t
                
            rt2 = _RPCTIME
            rt = rt2 - rt
            
            dct = dict(time=t, rpctime=rt, efftime=t-rt)
            logger.info(message(func, *args, **kw), extra=dct)
            
            return res
        
        func_wrapper.__name__ = func.__name__
        func_wrapper.__doc__ = func.__doc__
        func_wrapper.__dict__ = func.__dict__.copy()
        func_wrapper.__module__ = func.__module__
        return func_wrapper
        
    return wrapper

if not prof_on:
    def profile(*args, **kw):
        return lambda f: f

class ProfileViewer(object):
    
    def statfiles(self):
        if not exists(prof_dir): return []
        return [f for f in os.listdir(prof_dir) if f.endswith(".log")]
    
    def index(self):
        return """<html>
        <head><title>Profile Information</title></head>
        <frameset cols='200, 1*'>
            <frame src='menu' />
            <frame name='main' src=''/>
        </frameset>
        </html>
        """
    index.exposed = True
    
    def menu(self):
        yield "<h2>Profiling logs</h2>"
        yield "<p>Click on one of the log below to see profiling data.</p>"
        runs = self.statfiles()
        runs.sort()
        for i in runs:
            yield "<a href='report?filename=%s' target='main'>%s</a><br />" % (i, i[:-4])
        pass
    menu.exposed = True
    
    def report(self, filename):
        import cherrypy
        cherrypy.response.headers['Content-Type'] = 'text/plain'
        yield "%10s %10s %10s\n%s\n" % ('time', 'rpctime', 'efftime', ('-' * 33))
        yield open(join(prof_dir, filename)).read()
    report.exposed = True

profile.profiler = ProfileViewer()

__builtins__['profile'] = profile




import os

from itertools import izip, chain, imap

import cherrypy

from base import Widget
from utils import OrderedSet
from utils import Enum


locations = Enum(["head", "bodytop", "bodybottom"])


class Resource(Widget):

    location = locations.head

    @property
    def name(self):
        return None


class Link(Resource):

    params = {
        'link': None,
    }

    filename = None
    modname = None

    def __init__(self, modname, filename, location=locations.head, **kw):
        super(Link, self).__init__(modname=modname, filename=filename, location=location, **kw)

        #TODO: consider the server.webpath and cherrypy mount point
        self.link = "/cp_widgets/%s/%s" % (self.modname, self.filename)

    def __eq__(self, other):
        return getattr(other, 'link', None) == self.link

    def __hash__(self):
        return hash(self.link)


class JSLink(Link):

    template = """\
    <script type="text/javascript" src="${link}" ${py.attrs(charset=charset, defer=defer)}></script>\
    """

    params = {
        'charset': 'The character encoding of the linked script',
        'defer': 'If true, browser may defer execution of the script'
    }

    charset = None
    defer = None


class CSSLink(Link):

    params = {
        'media': 'Specify the media attribute for the css link tag'
    }

    template = """\
    <link rel="stylesheet" type="text/css" href="${link}" ${py.attrs(media=media)}/>\
    """


class Source(Resource):

    params = {
        'src': 'The source text',
    }

    def __init__(self, src, location=locations.head, **kw):
        super(Source, self).__init__(src=src, location=location, **kw)

    def __hash__(self):
        return hash(self.src)

    def __eq__(self, other):
        return self.src == getattr(other, 'src', None)


class JSSource(Source):
    """A JavaScript source snippet."""

    template = """\
    <script type="text/javascript" ${py.attrs(defer=defer)}>
        ${src|n}
    </script>\
    """

    params = {
        'defer': 'If true, browser may defer execution of the script'
    }


class CSSSource(Source):
    """A CSS source snippet."""

    template = """\
    <style type="text/css" ${py.attrs(media=media)}>
        ${src|n}
    </style>\
    """

    params = {
        'src': 'The CSS source for the link',
        'media' : 'Specify the media the css source link is for'
    }

    media = "all"

def register_resource_directory(app, modulename, directory):
    """Set up an application wide static resource directory...
    """

    assert isinstance(app, cherrypy.Application), "Excepected cherrypy.Application"

    directory = os.path.abspath(directory)
    app.config.update({'/cp_widgets/%s' % modulename: {
        'tools.staticdir.on': True,
        'tools.staticdir.dir': directory
    }})





###############################################################################
#
# Copyright (C) 2007-TODAY Tiny ERP Pvt Ltd. All Rights Reserved.
#
# $Id$
#
# Developed by Tiny (http://openerp.com) and Axelor (http://axelor.com).
#
# The OpenERP web client is distributed under the "OpenERP Public License".
# It's based on Mozilla Public License Version (MPL) 1.1 with following
# restrictions:
#
# -   All names, links and logos of Tiny, Open ERP and Axelor must be
#     kept as in original distribution without any changes in all software
#     screens, especially in start-up page and the software header, even if
#     the application source code has been changed or updated or code has been
#     added.
#
# -   All distributions of the software must keep source code with OEPL.
#
# -   All integrations to any other software must keep source code with OEPL.
#
# If you need commercial licence to remove this kind of restriction please
# contact us.
#
# You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

import re

from openerp.tools import expose
from openerp.tools import redirect
from openerp.tools import validate
from openerp.tools import error_handler

import cherrypy

from openerp import rpc
from openerp import tools
from openerp import common

from openerp import widgets as tw

from openerp.controllers.base import SecuredController
from openerp.controllers.form import default_error_handler
from openerp.utils import TinyDict

from openerp import validators

import form

class Wizard(SecuredController):

    def execute(self, params):

        action = params.name
        model = params.model
        state = params.state
        datas = params.datas

        form = None
        buttons = []

        if model:
            action = model.replace('wizard.', '', 1)
        else:
            model = 'wizard.' + action

        params.name = action
        params.model = model
        params.view_mode = []

        if 'form' not in datas:
            datas['form'] = {}

        wiz_id = params.wiz_id or rpc.session.execute('wizard', 'create', action)

        while state != 'end':

            ctx = rpc.session.context.copy()
            ctx.update(params.context or {})

            res = rpc.session.execute('wizard', 'execute', wiz_id, datas, state, ctx)

            if 'datas' in res:
                datas['form'].update(res['datas'])
            else:
                res['datas'] = {}

            if res['type']=='form':
                
                fields = res['fields']
                form_values = {}

                for f in fields:
                    if 'value' in fields[f]:
                        form_values[f] = fields[f]['value']

                form_values.update(datas['form'])

                datas['form'] = form_values

                res['datas'].update(datas['form'])
                
                params.is_wizard = True
                params.view_mode = ['form']
                params.view_type = 'form'
                params.views = {'form': res}
                
                # keep track of datas and some other required information
                params.hidden_fields = [tw.form.Hidden(name='_terp_datas', default=ustr(datas)),
                                        tw.form.Hidden(name='_terp_state2', default=state),
                                        tw.form.Hidden(name='_terp_wiz_id', default=wiz_id)]
                
                form = tw.form_view.ViewForm(params, name="view_form", action="/wizard/action")

                buttons = res.get('state', [])
                buttons = [(b[0], re.sub('_(?!_)', '', b[1])) for b in buttons] # remove mnemonic
                params.state = state

                return dict(form=form, buttons=buttons)

            elif res['type']=='action':
                from openerp.controllers import actions

                act_res = actions.execute(res['action'], **datas)
                if act_res:
                    return act_res

                state = res['state']

            elif res['type']=='print':
                from openerp.controllers import actions

                datas['report_id'] = res.get('report_id', False)
                if res.get('get_id_from_action', False):
                    backup_ids = datas['ids']
                    datas['ids'] = datas['form']['ids']

                return actions.execute_report(res['report'], **datas)

            elif res['type']=='state':
                state = res['state']

        raise redirect('/wizard/end')

    @expose(template="templates/wizard.mako")
    def create(self, params, tg_errors=None):

        if tg_errors:
            form = cherrypy.request.terp_form
            buttons = cherrypy.request.terp_buttons
            return dict(form=form, buttons=buttons)

        return self.execute(params)

    @expose()
    def end(self, **kw):

        if 'wizard_parent_params' in cherrypy.session:
            frm = cherrypy.session['wizard_parent_form']
            params = cherrypy.session['wizard_parent_params']

            frm = eval('cherrypy.request.app.root' + frm.replace('/', '.'))
            return frm.create(params)

        return """<html>
            <head>
                <script type="text/javascript">
                    if (window.opener) {
                        window.opener.setTimeout("window.location.reload()", 0);
                        window.close();
                    } else {
                        window.location.href = "/";
                    }
                </script>
            </head>
            <body>
            </body>
        </html>
        """

    def get_validation_schema(self):

        kw = cherrypy.request.params
        params, datas = TinyDict.split(kw)

        params.state = params.state2

        cherrypy.request.terp_validators = {}

        res = self.execute(params)

        form = res['form']
        buttons = res.get('buttons', [])

        cherrypy.request.terp_form = form
        cherrypy.request.terp_buttons = buttons

        vals = cherrypy.request.terp_validators
        keys = vals.keys()
        for k in keys:
            if k not in kw:
                vals.pop(k)

        form.validator = validators.Schema(**vals)
        return form

    @expose()
    @validate(form=get_validation_schema)
    @error_handler(default_error_handler)
    def report(self, tg_errors=None, tg_exceptions=None, **kw):

        if tg_exceptions:
            raise tg_exceptions

        params, datas = TinyDict.split(kw)
        params.datas['form'].update(datas)

        if tg_errors:
            return self.create(params, tg_errors=tg_errors)

        return self.execute(params)

    @expose()
    @validate(form=get_validation_schema)
    @error_handler(default_error_handler)
    def action(self, tg_errors=None, tg_exceptions=None, **kw):

        if tg_exceptions:
            raise tg_exceptions

        params, datas = TinyDict.split(kw)
        params.datas['form'].update(datas)

        return self.create(params, tg_errors=tg_errors)

# vim: ts=4 sts=4 sw=4 si et

